# -*- coding: utf-8 -*-
import json
import logging
import re
from bisect import bisect_left
from datetime import date, datetime, time, timedelta, timezone

from nectar.account import Account
from nectar.amount import Amount
from nectar.constants import STEEM_100_PERCENT, STEEM_VOTE_REGENERATION_SECONDS
from nectar.instance import shared_blockchain_instance
from nectar.utils import (
    addTzInfo,
    formatTimeString,
    parse_time,
    reputation_to_score,
)
from nectar.vote import Vote

log = logging.getLogger(__name__)


class AccountSnapshot(list):
    """This class allows to easily access Account history

    :param str account_name: Name of the account
    :param Steem blockchain_instance: Steem
           instance
    """

    def __init__(self, account, account_history=[], blockchain_instance=None, **kwargs):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        self.account = Account(account, blockchain_instance=self.blockchain)
        self.reset()
        super(AccountSnapshot, self).__init__(account_history)

    def reset(self):
        """Resets the arrays not the stored account history"""
        self.own_vests = [
            Amount(0, self.blockchain.vest_token_symbol, blockchain_instance=self.blockchain)
        ]
        self.own_steem = [
            Amount(0, self.blockchain.token_symbol, blockchain_instance=self.blockchain)
        ]
        self.own_sbd = [
            Amount(0, self.blockchain.backed_token_symbol, blockchain_instance=self.blockchain)
        ]
        self.delegated_vests_in = [{}]
        self.delegated_vests_out = [{}]
        self.timestamps = [addTzInfo(datetime(1970, 1, 1, 0, 0, 0, 0))]
        self.ops_statistics = {}
        for key in self.blockchain.get_operation_names():
            self.ops_statistics[key] = 0
        self.reward_timestamps = []
        self.author_rewards = []
        self.curation_rewards = []
        self.curation_per_1000_SP_timestamp = []
        self.curation_per_1000_SP = []
        self.out_vote_timestamp = []
        self.out_vote_weight = []
        self.in_vote_timestamp = []
        self.in_vote_weight = []
        self.in_vote_rep = []
        self.in_vote_rshares = []
        self.vp = []
        self.vp_timestamp = []
        self.downvote_vp = []
        self.downvote_vp_timestamp = []
        self.rep = []
        self.rep_timestamp = []

    def search(self, search_str, start=None, stop=None, use_block_num=True):
        """Returns ops in the given range"""
        ops = []
        if start is not None:
            start = addTzInfo(start)
        if stop is not None:
            stop = addTzInfo(stop)
        for op in self:
            if use_block_num and start is not None and isinstance(start, int):
                if op["block"] < start:
                    continue
            elif not use_block_num and start is not None and isinstance(start, int):
                if op["index"] < start:
                    continue
            elif start is not None and isinstance(start, (datetime, date, time)):
                if start > formatTimeString(op["timestamp"]):
                    continue
            if use_block_num and stop is not None and isinstance(stop, int):
                if op["block"] > stop:
                    continue
            elif not use_block_num and stop is not None and isinstance(stop, int):
                if op["index"] > stop:
                    continue
            elif stop is not None and isinstance(stop, (datetime, date, time)):
                if stop < formatTimeString(op["timestamp"]):
                    continue
            op_string = json.dumps(list(op.values()))
            if re.search(search_str, op_string):
                ops.append(op)
        return ops

    def get_ops(self, start=None, stop=None, use_block_num=True, only_ops=[], exclude_ops=[]):
        """Returns ops in the given range"""
        if start is not None:
            start = addTzInfo(start)
        if stop is not None:
            stop = addTzInfo(stop)
        for op in self:
            if use_block_num and start is not None and isinstance(start, int):
                if op["block"] < start:
                    continue
            elif not use_block_num and start is not None and isinstance(start, int):
                if op["index"] < start:
                    continue
            elif start is not None and isinstance(start, (datetime, date, time)):
                if start > formatTimeString(op["timestamp"]):
                    continue
            if use_block_num and stop is not None and isinstance(stop, int):
                if op["block"] > stop:
                    continue
            elif not use_block_num and stop is not None and isinstance(stop, int):
                if op["index"] > stop:
                    continue
            elif stop is not None and isinstance(stop, (datetime, date, time)):
                if stop < formatTimeString(op["timestamp"]):
                    continue
            if exclude_ops and op["type"] in exclude_ops:
                continue
            if not only_ops or op["type"] in only_ops:
                yield op

    def get_data(self, timestamp=None, index=0):
        """Returns snapshot for given timestamp"""
        if timestamp is None:
            timestamp = datetime.now(timezone.utc)
        timestamp = addTzInfo(timestamp)
        # Find rightmost value less than x
        i = bisect_left(self.timestamps, timestamp)
        if i:
            index = i - 1
        else:
            return {}
        ts = self.timestamps[index]
        own = self.own_vests[index]
        din = self.delegated_vests_in[index]
        dout = self.delegated_vests_out[index]
        steem = self.own_steem[index]
        sbd = self.own_sbd[index]
        sum_in = sum([din[key].amount for key in din])
        sum_out = sum([dout[key].amount for key in dout])
        from nectar import Steem

        if isinstance(self.blockchain, Steem):
            sp_in = self.blockchain.vests_to_sp(sum_in, timestamp=ts)
            sp_out = self.blockchain.vests_to_sp(sum_out, timestamp=ts)
            sp_own = self.blockchain.vests_to_sp(own, timestamp=ts)
        else:
            sp_in = self.blockchain.vests_to_hp(sum_in, timestamp=ts)
            sp_out = self.blockchain.vests_to_hp(sum_out, timestamp=ts)
            sp_own = self.blockchain.vests_to_hp(own, timestamp=ts)
        sp_eff = sp_own + sp_in - sp_out
        return {
            "timestamp": ts,
            "vests": own,
            "delegated_vests_in": din,
            "delegated_vests_out": dout,
            "sp_own": sp_own,
            "sp_eff": sp_eff,
            "steem": steem,
            "sbd": sbd,
            "index": index,
        }

    def get_account_history(self, start=None, stop=None, use_block_num=True):
        """Uses account history to fetch all related ops

        :param start: start number/date of transactions to
            return (*optional*)
        :type start: int, datetime
        :param stop: stop number/date of transactions to
            return (*optional*)
        :type stop: int, datetime
        :param bool use_block_num: if true, start and stop are block numbers,
            otherwise virtual OP count numbers.

        """
        super(AccountSnapshot, self).__init__(
            [h for h in self.account.history(start=start, stop=stop, use_block_num=use_block_num)]
        )

    def update_rewards(self, timestamp, curation_reward, author_vests, author_steem, author_sbd):
        self.reward_timestamps.append(timestamp)
        self.curation_rewards.append(curation_reward)
        self.author_rewards.append(
            {"vests": author_vests, "steem": author_steem, "sbd": author_sbd}
        )

    def update_out_vote(self, timestamp, weight):
        self.out_vote_timestamp.append(timestamp)
        self.out_vote_weight.append(weight)

    def update_in_vote(self, timestamp, weight, op):
        v = Vote(op)
        try:
            v.refresh()
            self.in_vote_timestamp.append(timestamp)
            self.in_vote_weight.append(weight)
            self.in_vote_rep.append(int(v["reputation"]))
            self.in_vote_rshares.append(int(v["rshares"]))
        except Exception:
            print("Could not find: %s" % v)
            return

    def update(self, timestamp, own, delegated_in=None, delegated_out=None, steem=0, sbd=0):
        """Updates the internal state arrays

        :param datetime timestamp: datetime of the update
        :param own: vests
        :type own: amount.Amount, float
        :param dict delegated_in: Incoming delegation
        :param dict delegated_out: Outgoing delegation
        :param steem: steem
        :type steem: amount.Amount, float
        :param sbd: sbd
        :type sbd: amount.Amount, float

        """
        self.timestamps.append(timestamp - timedelta(seconds=1))
        self.own_vests.append(self.own_vests[-1])
        self.own_steem.append(self.own_steem[-1])
        self.own_sbd.append(self.own_sbd[-1])
        self.delegated_vests_in.append(self.delegated_vests_in[-1])
        self.delegated_vests_out.append(self.delegated_vests_out[-1])

        self.timestamps.append(timestamp)
        self.own_vests.append(self.own_vests[-1] + own)
        self.own_steem.append(self.own_steem[-1] + steem)
        self.own_sbd.append(self.own_sbd[-1] + sbd)

        new_deleg = dict(self.delegated_vests_in[-1])
        if delegated_in is not None and delegated_in:
            if delegated_in["amount"] == 0:
                del new_deleg[delegated_in["account"]]
            else:
                new_deleg[delegated_in["account"]] = delegated_in["amount"]
        self.delegated_vests_in.append(new_deleg)

        new_deleg = dict(self.delegated_vests_out[-1])
        if delegated_out is not None and delegated_out:
            if delegated_out["account"] is None:
                # return_vesting_delegation
                for delegatee in new_deleg:
                    if new_deleg[delegatee]["amount"] == delegated_out["amount"]:
                        del new_deleg[delegatee]
                        break

            elif delegated_out["amount"] != 0:
                # new or updated non-zero delegation
                new_deleg[delegated_out["account"]] = delegated_out["amount"]
                # TODO
                # skip undelegations here, wait for 'return_vesting_delegation'
                # del new_deleg[delegated_out['account']]

        self.delegated_vests_out.append(new_deleg)

    def build(
        self,
        only_ops=[],
        exclude_ops=[],
        enable_rewards=False,
        enable_out_votes=False,
        enable_in_votes=False,
    ):
        """Builds the account history based on all account operations

        :param array only_ops: Limit generator by these
            operations (*optional*)
        :param array exclude_ops: Exclude these operations from
            generator (*optional*)

        """
        if len(self.timestamps) > 0:
            start_timestamp = self.timestamps[-1]
        else:
            start_timestamp = None
        for op in sorted(self, key=lambda k: k["timestamp"]):
            ts = parse_time(op["timestamp"])
            if start_timestamp is not None and start_timestamp > ts:
                continue
            if op["type"] in exclude_ops:
                continue
            if len(only_ops) > 0 and op["type"] not in only_ops:
                continue
            self.ops_statistics[op["type"]] += 1
            self.parse_op(
                op,
                only_ops=only_ops,
                enable_rewards=enable_rewards,
                enable_out_votes=enable_out_votes,
                enable_in_votes=enable_in_votes,
            )

    def parse_op(
        self, op, only_ops=[], enable_rewards=False, enable_out_votes=False, enable_in_votes=False
    ):
        """Parse account history operation"""
        ts = parse_time(op["timestamp"])

        if op["type"] == "account_create":
            fee_steem = Amount(op["fee"], blockchain_instance=self.blockchain).amount
            fee_vests = self.blockchain.sp_to_vests(
                Amount(op["fee"], blockchain_instance=self.blockchain).amount, timestamp=ts
            )
            if op["new_account_name"] == self.account["name"]:
                self.update(ts, fee_vests, 0, 0)
                return
            if op["creator"] == self.account["name"]:
                self.update(ts, 0, 0, 0, fee_steem * (-1), 0)
                return

        elif op["type"] == "account_create_with_delegation":
            fee_steem = Amount(op["fee"], blockchain_instance=self.blockchain).amount
            from nectar import Steem

            if isinstance(self.blockchain, Steem):
                fee_vests = self.blockchain.sp_to_vests(
                    Amount(op["fee"], blockchain_instance=self.blockchain).amount, timestamp=ts
                )
            else:
                fee_vests = self.blockchain.hp_to_vests(
                    Amount(op["fee"], blockchain_instance=self.blockchain).amount, timestamp=ts
                )
            if op["new_account_name"] == self.account["name"]:
                if Amount(op["delegation"], blockchain_instance=self.blockchain).amount > 0:
                    delegation = {
                        "account": op["creator"],
                        "amount": Amount(op["delegation"], blockchain_instance=self.blockchain),
                    }
                else:
                    delegation = None
                self.update(ts, fee_vests, delegation, 0)
                return

            if op["creator"] == self.account["name"]:
                delegation = {
                    "account": op["new_account_name"],
                    "amount": Amount(op["delegation"], blockchain_instance=self.blockchain),
                }
                self.update(ts, 0, 0, delegation, fee_steem * (-1), 0)
                return

        elif op["type"] == "delegate_vesting_shares":
            vests = Amount(op["vesting_shares"], blockchain_instance=self.blockchain)
            if op["delegator"] == self.account["name"]:
                delegation = {"account": op["delegatee"], "amount": vests}
                self.update(ts, 0, 0, delegation)
                return
            if op["delegatee"] == self.account["name"]:
                delegation = {"account": op["delegator"], "amount": vests}
                self.update(ts, 0, delegation, 0)
                return

        elif op["type"] == "transfer":
            amount = Amount(op["amount"], blockchain_instance=self.blockchain)
            if op["from"] == self.account["name"]:
                if amount.symbol == self.blockchain.blockchain_symbol:
                    self.update(ts, 0, 0, 0, amount * (-1), 0)
                elif amount.symbol == self.blockchain.backed_token_symbol:
                    self.update(ts, 0, 0, 0, 0, amount * (-1))
            if op["to"] == self.account["name"]:
                if amount.symbol == self.blockchain.blockchain_symbol:
                    self.update(ts, 0, 0, 0, amount, 0)
                elif amount.symbol == self.blockchain.backed_token_symbol:
                    self.update(ts, 0, 0, 0, 0, amount)
            return

        elif op["type"] == "fill_order":
            current_pays = Amount(op["current_pays"], blockchain_instance=self.blockchain)
            open_pays = Amount(op["open_pays"], blockchain_instance=self.blockchain)
            if op["current_owner"] == self.account["name"]:
                if current_pays.symbol == self.blockchain.token_symbol:
                    self.update(ts, 0, 0, 0, current_pays * (-1), open_pays)
                elif current_pays.symbol == self.blockchain.backed_token_symbol:
                    self.update(ts, 0, 0, 0, open_pays, current_pays * (-1))
            if op["open_owner"] == self.account["name"]:
                if current_pays.symbol == self.blockchain.token_symbol:
                    self.update(ts, 0, 0, 0, current_pays, open_pays * (-1))
                elif current_pays.symbol == self.blockchain.backed_token_symbol:
                    self.update(ts, 0, 0, 0, open_pays * (-1), current_pays)
            return

        elif op["type"] == "transfer_to_vesting":
            steem = Amount(op["amount"], blockchain_instance=self.blockchain)
            from nectar import Steem

            if isinstance(self.blockchain, Steem):
                vests = self.blockchain.sp_to_vests(steem.amount, timestamp=ts)
            else:
                vests = self.blockchain.hp_to_vests(steem.amount, timestamp=ts)
            if op["from"] == self.account["name"] and op["to"] == self.account["name"]:
                self.update(ts, vests, 0, 0, steem * (-1), 0)  # power up from and to given account
            elif op["from"] != self.account["name"] and op["to"] == self.account["name"]:
                self.update(ts, vests, 0, 0, 0, 0)  # power up from another account
            else:  # op['from'] == self.account["name"] and op['to'] != self.account["name"]
                self.update(ts, 0, 0, 0, steem * (-1), 0)  # power up to another account
            return

        elif op["type"] == "fill_vesting_withdraw":
            vests = Amount(op["withdrawn"], blockchain_instance=self.blockchain)
            self.update(ts, vests * (-1), 0, 0)
            return

        elif op["type"] == "return_vesting_delegation":
            delegation = {
                "account": None,
                "amount": Amount(op["vesting_shares"], blockchain_instance=self.blockchain),
            }
            self.update(ts, 0, 0, delegation)
            return

        elif op["type"] == "claim_reward_balance":
            vests = Amount(op["reward_vests"], blockchain_instance=self.blockchain)
            steem = Amount(op["reward_steem"], blockchain_instance=self.blockchain)
            sbd = Amount(op["reward_sbd"], blockchain_instance=self.blockchain)
            self.update(ts, vests, 0, 0, steem, sbd)
            return

        elif op["type"] == "curation_reward":
            if "curation_reward" in only_ops or enable_rewards:
                vests = Amount(op["reward"], blockchain_instance=self.blockchain)
            if "curation_reward" in only_ops:
                self.update(ts, vests, 0, 0)
            if enable_rewards:
                self.update_rewards(ts, vests, 0, 0, 0)
            return

        elif op["type"] == "author_reward":
            if "author_reward" in only_ops or enable_rewards:
                vests = Amount(op["vesting_payout"], blockchain_instance=self.blockchain)
                steem = Amount(op["steem_payout"], blockchain_instance=self.blockchain)
                sbd = Amount(op["sbd_payout"], blockchain_instance=self.blockchain)
            if "author_reward" in only_ops:
                self.update(ts, vests, 0, 0, steem, sbd)
            if enable_rewards:
                self.update_rewards(ts, 0, vests, steem, sbd)
            return

        elif op["type"] == "producer_reward":
            vests = Amount(op["vesting_shares"], blockchain_instance=self.blockchain)
            self.update(ts, vests, 0, 0)
            return

        elif op["type"] == "comment_benefactor_reward":
            if op["benefactor"] == self.account["name"]:
                if "reward" in op:
                    vests = Amount(op["reward"], blockchain_instance=self.blockchain)
                    self.update(ts, vests, 0, 0)
                else:
                    vests = Amount(op["vesting_payout"], blockchain_instance=self.blockchain)
                    steem = Amount(op["steem_payout"], blockchain_instance=self.blockchain)
                    sbd = Amount(op["sbd_payout"], blockchain_instance=self.blockchain)
                    self.update(ts, vests, 0, 0, steem, sbd)
                return
            else:
                return

        elif op["type"] == "fill_convert_request":
            amount_in = Amount(op["amount_in"], blockchain_instance=self.blockchain)
            amount_out = Amount(op["amount_out"], blockchain_instance=self.blockchain)
            if op["owner"] == self.account["name"]:
                self.update(ts, 0, 0, 0, amount_out, amount_in * (-1))
            return

        elif op["type"] == "interest":
            interest = Amount(op["interest"], blockchain_instance=self.blockchain)
            self.update(ts, 0, 0, 0, 0, interest)
            return

        elif op["type"] == "vote":
            if "vote" in only_ops or enable_out_votes:
                weight = int(op["weight"])
                if op["voter"] == self.account["name"]:
                    self.update_out_vote(ts, weight)
            if "vote" in only_ops or enable_in_votes and op["author"] == self.account["name"]:
                weight = int(op["weight"])
                self.update_in_vote(ts, weight, op)
            return

        elif op["type"] == "hardfork_hive":
            vests = Amount(op["vests_converted"])
            hbd = Amount(op["steem_transferred"])
            hive = Amount(op["sbd_transferred"])
            self.update(ts, vests * (-1), 0, 0, hive * (-1), hbd * (-1))

        elif op["type"] in [
            "comment",
            "feed_publish",
            "shutdown_witness",
            "account_witness_vote",
            "witness_update",
            "custom_json",
            "limit_order_create",
            "account_update",
            "account_witness_proxy",
            "limit_order_cancel",
            "comment_options",
            "delete_comment",
            "interest",
            "recover_account",
            "pow",
            "fill_convert_request",
            "convert",
            "request_account_recovery",
            "update_proposal_votes",
        ]:
            return

    def build_sp_arrays(self):
        """Builds the own_sp and eff_sp array"""
        self.own_sp = []
        self.eff_sp = []
        for ts, own, din, dout in zip(
            self.timestamps, self.own_vests, self.delegated_vests_in, self.delegated_vests_out
        ):
            sum_in = sum([din[key].amount for key in din])
            sum_out = sum([dout[key].amount for key in dout])
            from nectar import Steem

            if isinstance(self.blockchain, Steem):
                sp_in = self.blockchain.vests_to_sp(sum_in, timestamp=ts)
                sp_out = self.blockchain.vests_to_sp(sum_out, timestamp=ts)
                sp_own = self.blockchain.vests_to_sp(own, timestamp=ts)
            else:
                sp_in = self.blockchain.vests_to_hp(sum_in, timestamp=ts)
                sp_out = self.blockchain.vests_to_hp(sum_out, timestamp=ts)
                sp_own = self.blockchain.vests_to_hp(own, timestamp=ts)

            sp_eff = sp_own + sp_in - sp_out
            self.own_sp.append(sp_own)
            self.eff_sp.append(sp_eff)

    def build_rep_arrays(self):
        """Build reputation arrays"""
        self.rep_timestamp = [self.timestamps[1]]
        self.rep = [reputation_to_score(0)]
        current_reputation = 0
        for ts, rshares, rep in zip(self.in_vote_timestamp, self.in_vote_rshares, self.in_vote_rep):
            if rep > 0:
                if rshares > 0 or (rshares < 0 and rep > current_reputation):
                    current_reputation += rshares >> 6
            self.rep.append(reputation_to_score(current_reputation))
            self.rep_timestamp.append(ts)

    def build_vp_arrays(self):
        """Build vote power arrays"""
        self.vp_timestamp = [self.timestamps[1]]
        self.vp = [STEEM_100_PERCENT]
        HF_21 = datetime(2019, 8, 27, 15, tzinfo=timezone.utc)
        if self.timestamps[1] > HF_21:
            self.downvote_vp_timestamp = [self.timestamps[1]]
        else:
            self.downvote_vp_timestamp = [HF_21]
        self.downvote_vp = [STEEM_100_PERCENT]

        for ts, weight in zip(self.out_vote_timestamp, self.out_vote_weight):
            regenerated_vp = 0
            if ts > HF_21 and weight < 0:
                self.downvote_vp.append(self.downvote_vp[-1])
                if self.downvote_vp[-1] < STEEM_100_PERCENT:
                    regenerated_vp = (
                        ((ts - self.downvote_vp_timestamp[-1]).total_seconds())
                        * STEEM_100_PERCENT
                        / STEEM_VOTE_REGENERATION_SECONDS
                    )
                    self.downvote_vp[-1] += int(regenerated_vp)

                if self.downvote_vp[-1] > STEEM_100_PERCENT:
                    self.downvote_vp[-1] = STEEM_100_PERCENT
                    recharge_time = self.account.get_manabar_recharge_timedelta(
                        {"current_mana_pct": self.downvote_vp[-2] / 100}
                    )
                    # Add full downvote VP once fully charged
                    self.downvote_vp_timestamp.append(
                        self.downvote_vp_timestamp[-1] + recharge_time
                    )
                    self.downvote_vp.append(STEEM_100_PERCENT)

                # Add charged downvote VP just before new Vote
                self.downvote_vp_timestamp.append(ts - timedelta(seconds=1))
                self.downvote_vp.append(
                    min([STEEM_100_PERCENT, self.downvote_vp[-1] + regenerated_vp])
                )

                self.downvote_vp[-1] -= (
                    self.blockchain._calc_resulting_vote(STEEM_100_PERCENT, weight) * 4
                )
                # Downvote mana pool is 1/4th of the upvote mana pool, so it gets drained 4 times as quick
                if self.downvote_vp[-1] < 0:
                    # There's most likely a better solution to this that what I did here
                    self.vp.append(self.vp[-1])

                    if self.vp[-1] < STEEM_100_PERCENT:
                        regenerated_vp = (
                            ((ts - self.vp_timestamp[-1]).total_seconds())
                            * STEEM_100_PERCENT
                            / STEEM_VOTE_REGENERATION_SECONDS
                        )
                        self.vp[-1] += int(regenerated_vp)

                    if self.vp[-1] > STEEM_100_PERCENT:
                        self.vp[-1] = STEEM_100_PERCENT
                        recharge_time = self.account.get_manabar_recharge_timedelta(
                            {"current_mana_pct": self.vp[-2] / 100}
                        )
                        # Add full VP once fully charged
                        self.vp_timestamp.append(self.vp_timestamp[-1] + recharge_time)
                        self.vp.append(STEEM_100_PERCENT)
                    if self.vp[-1] == STEEM_100_PERCENT and ts - self.vp_timestamp[-1] > timedelta(
                        seconds=1
                    ):
                        # Add charged VP just before new Vote
                        self.vp_timestamp.append(ts - timedelta(seconds=1))
                        self.vp.append(min([STEEM_100_PERCENT, self.vp[-1] + regenerated_vp]))
                    self.vp[-1] += self.downvote_vp[-1] / 4
                    if self.vp[-1] < 0:
                        self.vp[-1] = 0

                    self.vp_timestamp.append(ts)
                    self.downvote_vp[-1] = 0
                self.downvote_vp_timestamp.append(ts)

            else:
                self.vp.append(self.vp[-1])

                if self.vp[-1] < STEEM_100_PERCENT:
                    regenerated_vp = (
                        ((ts - self.vp_timestamp[-1]).total_seconds())
                        * STEEM_100_PERCENT
                        / STEEM_VOTE_REGENERATION_SECONDS
                    )
                    self.vp[-1] += int(regenerated_vp)

                if self.vp[-1] > STEEM_100_PERCENT:
                    self.vp[-1] = STEEM_100_PERCENT
                    recharge_time = self.account.get_manabar_recharge_timedelta(
                        {"current_mana_pct": self.vp[-2] / 100}
                    )
                    # Add full VP once fully charged
                    self.vp_timestamp.append(self.vp_timestamp[-1] + recharge_time)
                    self.vp.append(STEEM_100_PERCENT)
                if self.vp[-1] == STEEM_100_PERCENT and ts - self.vp_timestamp[-1] > timedelta(
                    seconds=1
                ):
                    # Add charged VP just before new Vote
                    self.vp_timestamp.append(ts - timedelta(seconds=1))
                    self.vp.append(min([STEEM_100_PERCENT, self.vp[-1] + regenerated_vp]))
                self.vp[-1] -= self.blockchain._calc_resulting_vote(self.vp[-1], weight)
                if self.vp[-1] < 0:
                    self.vp[-1] = 0

                self.vp_timestamp.append(ts)

        if self.account.get_voting_power() == 100:
            self.vp.append(10000)
            recharge_time = self.account.get_manabar_recharge_timedelta(
                {"current_mana_pct": self.vp[-2] / 100}
            )
            self.vp_timestamp.append(self.vp_timestamp[-1] + recharge_time)

        if self.account.get_downvoting_power() == 100:
            self.downvote_vp.append(10000)
            recharge_time = self.account.get_manabar_recharge_timedelta(
                {"current_mana_pct": self.downvote_vp[-2] / 100}
            )
            self.downvote_vp_timestamp.append(self.vp_timestamp[-1] + recharge_time)

        self.vp.append(self.account.get_voting_power() * 100)
        self.downvote_vp.append(self.account.get_downvoting_power() * 100)
        self.downvote_vp_timestamp.append(datetime.now(timezone.utc))
        self.vp_timestamp.append(datetime.now(timezone.utc))

    def build_curation_arrays(self, end_date=None, sum_days=7):
        """Build curation arrays"""
        self.curation_per_1000_SP_timestamp = []
        self.curation_per_1000_SP = []
        if sum_days <= 0:
            raise ValueError("sum_days must be greater than 0")
        index = 0
        curation_sum = 0
        days = (self.reward_timestamps[-1] - self.reward_timestamps[0]).days // sum_days * sum_days
        if end_date is None:
            end_date = self.reward_timestamps[-1] - timedelta(days=days)
        for ts, vests in zip(self.reward_timestamps, self.curation_rewards):
            if vests == 0:
                continue
            from nectar import Steem

            if isinstance(self.blockchain, Steem):
                sp = self.blockchain.vests_to_sp(vests, timestamp=ts)
            else:
                sp = self.blockchain.vests_to_hp(vests, timestamp=ts)
            data = self.get_data(timestamp=ts, index=index)
            index = data["index"]
            if "sp_eff" in data and data["sp_eff"] > 0:
                curation_1k_sp = sp / data["sp_eff"] * 1000 / sum_days * 7
            else:
                curation_1k_sp = 0
            if ts < end_date:
                curation_sum += curation_1k_sp
            else:
                self.curation_per_1000_SP_timestamp.append(end_date)
                self.curation_per_1000_SP.append(curation_sum)
                end_date = end_date + timedelta(days=sum_days)
                curation_sum = 0

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, str(self.account["name"]))



# -*- coding: utf-8 -*-
import unittest

from nectar import Steem
from nectar.blockchain import Blockchain
from nectar.instance import set_shared_steem_instance

from .nodes import get_hive_nodes


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.bts = Steem(
            node=get_hive_nodes(),
            nobroadcast=True,
            timeout=30,
            num_retries=30,
        )
        # from getpass import getpass
        # self.bts.wallet.unlock(getpass())
        set_shared_steem_instance(cls.bts)
        cls.bts.set_default_account("test")

        b = Blockchain(steem_instance=cls.bts)
        num = b.get_current_block_num()
        # num = 23346630
        cls.start = num - 25
        cls.stop = num
        # cls.N_transfer = 121
        # cls.N_vote = 2825

    def test_block_threading(self):
        bts = self.bts
        b = Blockchain(steem_instance=bts)
        blocks_no_threading = []
        for block in b.blocks(start=self.start, stop=self.stop, threading=False, thread_num=8):
            blocks_no_threading.append(block)

        for n in range(5):
            blocks = []
            for block in b.blocks(start=self.start, stop=self.stop, threading=True, thread_num=8):
                blocks.append(block)

            for i in range(min(len(blocks), len(blocks_no_threading))):
                self.assertEqual(blocks[i]["block_id"], blocks_no_threading[i]["block_id"])
            self.assertEqual(len(blocks_no_threading), len(blocks))

    def test_stream_threading(self):
        bts = self.bts
        b = Blockchain(steem_instance=bts)

        ops_stream_no_threading = []
        opNames = ["transfer", "vote"]

        block_num_list2 = []
        for op in b.stream(opNames=opNames, start=self.start, stop=self.stop, threading=False):
            ops_stream_no_threading.append(op)
            if op["block_num"] not in block_num_list2:
                block_num_list2.append(op["block_num"])
        for n in range(5):
            ops_stream = []
            block_num_list = []
            for op in b.stream(
                opNames=opNames, start=self.start, stop=self.stop, threading=True, thread_num=8
            ):
                ops_stream.append(op)
                if op["block_num"] not in block_num_list:
                    block_num_list.append(op["block_num"])

            self.assertEqual(ops_stream[0]["block_num"], ops_stream_no_threading[0]["block_num"])
            self.assertEqual(ops_stream[-1]["block_num"], ops_stream_no_threading[-1]["block_num"])
            self.assertEqual(len(ops_stream_no_threading), len(ops_stream))

        self.assertEqual(len(block_num_list), len(block_num_list2))
        for i in range(len(block_num_list)):
            self.assertEqual(block_num_list[i], block_num_list2[i])

    def test_stream_threading2(self):
        bts = self.bts
        b = Blockchain(steem_instance=bts)

        ops_stream = []
        start_block = 25097000
        stop_block = 25097100
        opNames = ["account_create", "custom_json"]
        for op in b.stream(
            start=int(start_block),
            stop=int(stop_block),
            opNames=opNames,
            threading=True,
            thread_num=8,
        ):
            ops_stream.append(op)
        self.assertTrue(ops_stream[0]["block_num"] >= start_block)
        self.assertTrue(ops_stream[-1]["block_num"] <= stop_block)
        op_stat = b.ops_statistics(start=start_block, stop=stop_block)
        self.assertEqual(op_stat["account_create"] + op_stat["custom_json"], len(ops_stream))



# -*- coding: utf-8 -*-
import time
import unittest

from nectar import Steem
from nectar.block import Block
from nectar.blockchain import Blockchain
from nectar.exceptions import BlockWaitTimeExceeded
from nectar.instance import set_shared_blockchain_instance
from nectarbase.signedtransactions import Signed_Transaction

from .nodes import get_hive_nodes

wif = "5KQwrPbwdL6PhXujxW37FSSQZ1JiwsST4cqQzDeyXtP79zkvFD3"


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.bts = Steem(
            node=get_hive_nodes(), nobroadcast=True, keys={"active": wif}, num_retries=10
        )
        b = Blockchain(blockchain_instance=cls.bts)
        num = b.get_current_block_num()
        cls.start = num - 5
        cls.stop = num

        # from getpass import getpass
        # self.bts.wallet.unlock(getpass())
        set_shared_blockchain_instance(cls.bts)

    def test_blockchain(self):
        bts = self.bts
        b = Blockchain(blockchain_instance=bts)
        num = b.get_current_block_num()
        self.assertTrue(num > 0)
        self.assertTrue(isinstance(num, int))
        block = b.get_current_block()
        self.assertTrue(isinstance(block, Block))
        # self.assertTrue(num <= block.identifier)
        block_time = b.block_time(block.identifier)
        self.assertEqual(block.time(), block_time)
        block_timestamp = b.block_timestamp(block.identifier)
        timestamp = int(time.mktime(block.time().timetuple()))
        self.assertEqual(block_timestamp, timestamp)

    def test_estimate_block_num(self):
        bts = self.bts
        b = Blockchain(blockchain_instance=bts)
        last_block = b.get_current_block()
        num = last_block.identifier
        old_block = Block(num - 60, blockchain_instance=bts)
        date = old_block.time()
        est_block_num = b.get_estimated_block_num(date, accurate=False)
        self.assertTrue((est_block_num - (old_block.identifier)) < 10)
        est_block_num = b.get_estimated_block_num(date, accurate=True)
        self.assertTrue((est_block_num - (old_block.identifier)) < 2)
        est_block_num = b.get_estimated_block_num(date, estimateForwards=True, accurate=True)
        self.assertTrue((est_block_num - (old_block.identifier)) < 2)
        est_block_num = b.get_estimated_block_num(date, estimateForwards=True, accurate=False)

    def test_get_account_count(self):
        b = Blockchain(blockchain_instance=self.bts)
        num = b.get_account_count()
        self.assertTrue(isinstance(num, int) and num > 0)

    def test_get_all_accounts(self):
        bts = self.bts
        b = Blockchain(blockchain_instance=bts)
        accounts = []
        limit = 200
        for acc in b.get_all_accounts(steps=100, limit=limit):
            accounts.append(acc)
        self.assertEqual(len(accounts), limit)
        self.assertEqual(len(set(accounts)), limit)

    def test_awaitTX(self):
        bts = self.bts
        b = Blockchain(blockchain_instance=bts)
        trans = {
            "ref_block_num": 3855,
            "ref_block_prefix": 1730859721,
            "expiration": "2018-03-09T06:21:06",
            "operations": [],
            "extensions": [],
            "signatures": [
                "2033a872a8ad33c7d5b946871e4c9cc8f08a5809258355fc909058eac83"
                "20ac2a872517a52b51522930d93dd2c1d5eb9f90b070f75f838c881ff29b11af98d6a1b"
            ],
        }
        with self.assertRaises(Exception):
            b.awaitTxConfirmation(trans)

    def test_stream(self):
        bts = self.bts
        start = self.start
        stop = self.stop
        b = Blockchain(blockchain_instance=bts)
        ops_stream = []
        opNames = ["transfer", "vote"]
        for op in b.stream(opNames=opNames, start=start, stop=stop):
            ops_stream.append(op)
        self.assertTrue(len(ops_stream) >= 0)

        ops_raw_stream = []
        opNames = ["transfer", "vote"]
        for op in b.stream(opNames=opNames, raw_ops=True, start=start, stop=stop):
            ops_raw_stream.append(op)
        self.assertTrue(len(ops_raw_stream) >= 0)

        only_ops_stream = []
        opNames = ["transfer", "vote"]
        for op in b.stream(opNames=opNames, start=start, stop=stop, only_ops=True):
            only_ops_stream.append(op)
        self.assertTrue(len(only_ops_stream) >= 0)

        only_ops_raw_stream = []
        opNames = ["transfer", "vote"]
        for op in b.stream(opNames=opNames, raw_ops=True, start=start, stop=stop, only_ops=True):
            only_ops_raw_stream.append(op)
        self.assertTrue(len(only_ops_raw_stream) >= 0)

        op_stat = b.ops_statistics(start=start, stop=stop)
        op_stat2 = {"transfer": 0, "vote": 0}
        for op in ops_stream:
            self.assertIn(op["type"], opNames)
            op_stat2[op["type"]] += 1
            self.assertTrue(op["block_num"] >= start)
            self.assertTrue(op["block_num"] <= stop)
        self.assertEqual(op_stat["transfer"], op_stat2["transfer"])
        self.assertEqual(op_stat["vote"], op_stat2["vote"])

        op_stat3 = {"transfer": 0, "vote": 0}
        for op in ops_raw_stream:
            self.assertIn(op["op"][0], opNames)
            op_stat3[op["op"][0]] += 1
            self.assertTrue(op["block_num"] >= start)
            self.assertTrue(op["block_num"] <= stop)
        self.assertEqual(op_stat["transfer"], op_stat3["transfer"])
        self.assertEqual(op_stat["vote"], op_stat3["vote"])

        op_stat5 = {"transfer": 0, "vote": 0}
        for op in only_ops_stream:
            self.assertIn(op["type"], opNames)
            op_stat5[op["type"]] += 1
            self.assertTrue(op["block_num"] >= start)
            self.assertTrue(op["block_num"] <= stop)
        self.assertEqual(op_stat["transfer"], op_stat5["transfer"])
        self.assertEqual(op_stat["vote"], op_stat5["vote"])

        op_stat6 = {"transfer": 0, "vote": 0}
        for op in only_ops_raw_stream:
            self.assertIn(op["op"][0], opNames)
            op_stat6[op["op"][0]] += 1
            self.assertTrue(op["block_num"] >= start)
            self.assertTrue(op["block_num"] <= stop)
        self.assertEqual(op_stat["transfer"], op_stat6["transfer"])
        self.assertEqual(op_stat["vote"], op_stat6["vote"])

        ops_blocks = []
        for op in b.blocks(start=start, stop=stop):
            ops_blocks.append(op)
        op_stat4 = {"transfer": 0, "vote": 0}
        self.assertTrue(len(ops_blocks) > 0)
        for block in ops_blocks:
            for tran in block["transactions"]:
                for op in tran["operations"]:
                    if isinstance(op, list) and op[0] in opNames:
                        op_stat4[op[0]] += 1
                    elif isinstance(op, dict):
                        op_type = op["type"]
                        if len(op_type) > 10 and op_type[len(op_type) - 10 :] == "_operation":
                            op_type = op_type[:-10]
                        if op_type in opNames:
                            op_stat4[op_type] += 1
            self.assertTrue(block.identifier >= start)
            self.assertTrue(block.identifier <= stop)
        self.assertEqual(op_stat["transfer"], op_stat4["transfer"])
        self.assertEqual(op_stat["vote"], op_stat4["vote"])

        ops_blocks = []
        for op in b.blocks():
            ops_blocks.append(op)
            break
        self.assertTrue(len(ops_blocks) == 1)

    def test_stream2(self):
        bts = self.bts
        b = Blockchain(blockchain_instance=bts)
        stop_block = b.get_current_block_num()
        start_block = stop_block - 10
        ops_stream = []
        for op in b.stream(start=start_block, stop=stop_block):
            ops_stream.append(op)
        self.assertTrue(len(ops_stream) > 0)

    def test_wait_for_and_get_block(self):
        bts = self.bts
        b = Blockchain(blockchain_instance=bts, max_block_wait_repetition=18)
        start_num = b.get_current_block_num()
        blocknum = start_num
        last_fetched_block_num = None
        for i in range(3):
            block = b.wait_for_and_get_block(blocknum)
            last_fetched_block_num = block.block_num
            blocknum = last_fetched_block_num + 1
        self.assertEqual(last_fetched_block_num, start_num + 2)

        b2 = Blockchain(blockchain_instance=bts, max_block_wait_repetition=1)
        with self.assertRaises(BlockWaitTimeExceeded):
            for i in range(300):
                block = b2.wait_for_and_get_block(blocknum)
                last_fetched_block_num = block.block_num
                blocknum = last_fetched_block_num + 2

    def test_hash_op(self):
        bts = self.bts
        b = Blockchain(blockchain_instance=bts)
        op1 = {
            "type": "vote_operation",
            "value": {
                "voter": "ubg",
                "author": "yesslife",
                "permlink": "steemit-sandwich-contest-week-25-2da-entry",
                "weight": 100,
            },
        }
        op2 = [
            "vote",
            {
                "voter": "ubg",
                "author": "yesslife",
                "permlink": "steemit-sandwich-contest-week-25-2da-entry",
                "weight": 100,
            },
        ]
        hash1 = b.hash_op(op1)
        hash2 = b.hash_op(op2)
        self.assertEqual(hash1, hash2)

    def test_signing_appbase(self):
        b = Blockchain(blockchain_instance=self.bts)
        st = None
        for block in b.blocks(start=25304468, stop=25304468):
            for trx in block.transactions:
                st = Signed_Transaction(trx.copy())
        self.assertTrue(st is not None)

    def test_get_account_reputations(self):
        b = Blockchain(blockchain_instance=self.bts)
        limit = 100  # get the first 100 account reputations
        reps_limit = list(b.get_account_reputations(limit=limit))
        self.assertTrue(len(reps_limit) == limit)
        for rep in reps_limit:  # expect format {'name': [str], 'reputation': [int]}
            self.assertTrue(isinstance(rep, dict))
            self.assertTrue("name" in rep and "reputation" in rep)
            self.assertTrue(isinstance(rep["name"], str))
            self.assertTrue(isinstance(rep["reputation"], int))

        first = reps_limit[0]["name"]
        last = reps_limit[-1]["name"]
        # get the same account reputations via start/stop constraints
        reps_constr = list(b.get_account_reputations(start=first, stop=last))
        self.assertTrue(len(reps_constr) >= limit)
        # The actual number of accounts may have increased and the
        # reputation values may be different between the two API
        # calls, but each account of the first call should be
        # contained in the second as well
        accounts = [rep["name"] for rep in reps_constr]
        for rep in reps_limit:
            self.assertTrue(rep["name"] in accounts)



# -*- coding: utf-8 -*-
import os
import random
import struct
from binascii import hexlify, unhexlify

from nectar.instance import shared_blockchain_instance
from nectar.version import version as __version__
from nectarbase import memo as BtsMemo
from nectargraphenebase.account import PrivateKey, PublicKey
from nectargraphenebase.base58 import base58decode, base58encode

from .account import Account
from .exceptions import MissingKeyError


class Memo(object):
    """Deals with Memos that are attached to a transfer

    :param Account from_account: Account that has sent the memo
    :param Account to_account: Account that has received the memo
    :param Steem blockchain_instance: Steem instance

    A memo is encrypted with a shared secret derived from a private key of
    the sender and a public key of the receiver. Due to the underlying
    mathematics, the same shared secret can be derived by the private key
    of the receiver and the public key of the sender. The encrypted message
    is perturbed by a nonce that is part of the transmitted message.

    .. code-block:: python

        from nectar.memo import Memo
        m = Memo("thecrazygm", "hive-nectar")
        m.unlock_wallet("secret")
        enc = (m.encrypt("test"))
        print(enc)
        >> {'message': '#DTpKcbxWqsETCRfjYGk9feERFa5nVBF8FaHfWPwUjyHBTgNhXGh4mN5TTG41nLhUcHtXfu7Hy3AwLrtWvo1ERUyAZaJjaEZNhHyoeDnrHdWChrzbccbANQmazgwjyxzEL', 'from': 'STM6MQBLaX9Q15CK3prXoWK4C6EqtsL7C4rqq1h6BQjxvfk9tuT3N', 'to': 'STM6sRudsxWpTZWxnpRkCDVD51RteiJnvJYCt5LiZAbVLfM1hJCQC'}
        print(m.decrypt(enc))
        >> foobar

    To decrypt a memo, simply use

    .. code-block:: python

        from nectar.memo import Memo
        m = Memo()
        m.unlock_wallet("secret")
        print(m.decrypt(op_data["memo"]))

    if ``op_data`` being the payload of a transfer operation.

    Memo Keys

    In Steem, memos are AES-256 encrypted with a shared secret between sender and
    receiver. It is derived from the memo private key of the sender and the memo
    public key of the receiver.

    In order for the receiver to decode the memo, the shared secret has to be
    derived from the receiver's private key and the senders public key.

    The memo public key is part of the account and can be retrieved with the
    `get_account` call:

    .. code-block:: js

        get_account <accountname>
        {
          [...]
          "options": {
            "memo_key": "GPH5TPTziKkLexhVKsQKtSpo4bAv5RnB8oXcG4sMHEwCcTf3r7dqE",
            [...]
          },
          [...]
        }

    while the memo private key can be dumped with `dump_private_keys`

    Memo Message

    The take the following form:

    .. code-block:: js

            {
              "from": "GPH5mgup8evDqMnT86L7scVebRYDC2fwAWmygPEUL43LjstQegYCC",
              "to": "GPH5Ar4j53kFWuEZQ9XhxbAja4YXMPJ2EnUg5QcrdeMFYUNMMNJbe",
              "nonce": "13043867485137706821",
              "message": "d55524c37320920844ca83bb20c8d008"
            }

    The fields `from` and `to` contain the memo public key of sender and receiver.
    The `nonce` is a random integer that is used for the seed of the AES encryption
    of the message.

    Encrypting a memo

    The high level memo class makes use of the nectar wallet to obtain keys
    for the corresponding accounts.

    .. code-block:: python

        from nectar.memo import Memo
        from nectar.account import Account

        memoObj = Memo(
            from_account=Account(from_account),
            to_account=Account(to_account)
        )
        encrypted_memo = memoObj.encrypt(memo)

    Decoding of a received memo

    .. code-block:: python

        from getpass import getpass
        from nectar.block import Block
        from nectar.memo import Memo

        # Obtain a transfer from the blockchain
        block = Block(23755086)                   # block
        transaction = block["transactions"][3]    # transactions
        op = transaction["operations"][0]         # operation
        op_id = op[0]                             # operation type
        op_data = op[1]                           # operation payload

        # Instantiate Memo for decoding
        memo = Memo()

        # Unlock wallet
        memo.unlock_wallet(getpass())

        # Decode memo
        # Raises exception if required keys not available in the wallet
        print(memo.decrypt(op_data["transfer"]))

    """

    def __init__(self, from_account=None, to_account=None, blockchain_instance=None, **kwargs):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()

        if to_account and len(to_account) < 51:
            self.to_account = Account(to_account, blockchain_instance=self.blockchain)
        elif to_account and len(to_account) >= 51:
            self.to_account = PublicKey(to_account)
        else:
            self.to_account = None
        if from_account and len(from_account) < 51:
            self.from_account = Account(from_account, blockchain_instance=self.blockchain)
        elif from_account and len(from_account) >= 51:
            self.from_account = PrivateKey(from_account)
        else:
            self.from_account = None

    def unlock_wallet(self, *args, **kwargs):
        """Unlock the library internal wallet"""
        self.blockchain.wallet.unlock(*args, **kwargs)

    def encrypt(self, memo, bts_encrypt=False, return_enc_memo_only=False, nonce=None):
        """Encrypt a memo

        :param str memo: clear text memo message
        :param bool return_enc_memo_only: When True, only the encoded memo is returned
        :param str nonce: when not set, a random string is generated and used
        :returns: encrypted memo
        :rtype: dict
        """
        if not memo:
            return None
        if nonce is None:
            nonce = str(random.getrandbits(64))
        if isinstance(self.from_account, Account):
            memo_wif = self.blockchain.wallet.getPrivateKeyForPublicKey(
                self.from_account["memo_key"]
            )
            memo_wif = PrivateKey(memo_wif)
        else:
            memo_wif = self.from_account
        if isinstance(self.to_account, Account):
            pubkey = self.to_account["memo_key"]
        else:
            pubkey = self.to_account
        if not memo_wif:
            raise MissingKeyError("Memo key for %s missing!" % self.from_account["name"])

        if not hasattr(self, "chain_prefix"):
            self.chain_prefix = self.blockchain.prefix

        if bts_encrypt:
            enc = BtsMemo.encode_memo_bts(
                PrivateKey(memo_wif), PublicKey(pubkey, prefix=self.chain_prefix), nonce, memo
            )

            return {
                "message": enc,
                "nonce": nonce,
                "from": str(PrivateKey(memo_wif).pubkey),
                "to": str(pubkey),
            }
        else:
            enc = BtsMemo.encode_memo(
                PrivateKey(memo_wif),
                PublicKey(pubkey, prefix=self.chain_prefix),
                nonce,
                memo,
                prefix=self.chain_prefix,
            )
            if return_enc_memo_only:
                return enc
            return {"message": enc, "from": str(PrivateKey(memo_wif).pubkey), "to": str(pubkey)}

    def encrypt_binary(self, infile, outfile, buffer_size=2048, nonce=None):
        """Encrypt a binary file

        :param str infile: input file name
        :param str outfile: output file name
        :param int buffer_size: write buffer size
        :param str nonce: when not set, a random string is generated and used
        """
        if not os.path.exists(infile):
            raise ValueError("%s does not exists!" % infile)

        if nonce is None:
            nonce = str(random.getrandbits(64))
        if isinstance(self.from_account, Account):
            memo_wif = self.blockchain.wallet.getPrivateKeyForPublicKey(
                self.from_account["memo_key"]
            )
        else:
            memo_wif = self.from_account
        if isinstance(self.to_account, Account):
            pubkey = self.to_account["memo_key"]
        else:
            pubkey = self.to_account
        if not memo_wif:
            raise MissingKeyError("Memo key for %s missing!" % self.from_account["name"])

        if not hasattr(self, "chain_prefix"):
            self.chain_prefix = self.blockchain.prefix

        file_size = os.path.getsize(infile)
        priv = PrivateKey(memo_wif)
        pub = PublicKey(pubkey, prefix=self.chain_prefix)
        enc = BtsMemo.encode_memo(
            priv, pub, nonce, "nectar/%s" % __version__, prefix=self.chain_prefix
        )
        enc = unhexlify(base58decode(enc[1:]))
        shared_secret = BtsMemo.get_shared_secret(priv, pub)
        aes, check = BtsMemo.init_aes(shared_secret, nonce)
        with open(outfile, "wb") as fout:
            fout.write(struct.pack("<Q", len(enc)))
            fout.write(enc)
            fout.write(struct.pack("<Q", file_size))
            with open(infile, "rb") as fin:
                while True:
                    data = fin.read(buffer_size)
                    n = len(data)
                    if n == 0:
                        break
                    elif n % 16 != 0:
                        data += b" " * (16 - n % 16)  # <- padded with spaces
                    encd = aes.encrypt(data)
                    fout.write(encd)

    def extract_decrypt_memo_data(self, memo):
        """Returns information about an encrypted memo"""
        from_key, to_key, nonce, check, cipher = BtsMemo.extract_memo_data(memo)
        return from_key, to_key, nonce

    def decrypt(self, memo):
        """Decrypt a memo

        :param str memo: encrypted memo message
        :returns: encrypted memo
        :rtype: str
        """
        if not memo:
            return None
        memo_wif = None
        # We first try to decode assuming we received the memo
        if isinstance(memo, dict) and "to" in memo and "from" in memo and "memo" in memo:
            memo_to = Account(memo["to"], blockchain_instance=self.blockchain)
            memo_from = Account(memo["from"], blockchain_instance=self.blockchain)
            message = memo["memo"]
        else:
            memo_to = self.to_account
            memo_from = self.from_account
            message = memo
        if isinstance(memo, dict) and "nonce" in memo:
            nonce = memo.get("nonce")
        else:
            nonce = ""

        if memo_to is None or memo_from is None:
            from_key, to_key, nonce, check, cipher = BtsMemo.extract_memo_data(message)
            try:
                memo_wif = self.blockchain.wallet.getPrivateKeyForPublicKey(str(to_key))
                pubkey = from_key
            except MissingKeyError:
                try:
                    # if that failed, we assume that we have sent the memo
                    memo_wif = self.blockchain.wallet.getPrivateKeyForPublicKey(str(from_key))
                    pubkey = to_key
                except MissingKeyError:
                    # if all fails, raise exception
                    raise MissingKeyError(
                        "Non of the required memo keys are installed!Need any of {}".format(
                            [str(to_key), str(from_key)]
                        )
                    )
        elif memo_to is not None and memo_from is not None and isinstance(memo_from, PrivateKey):
            memo_wif = memo_from
            pubkey = memo_to
        elif memo_to is not None and memo_from is not None and isinstance(memo_to, PrivateKey):
            memo_wif = memo_to
            pubkey = memo_from
        else:
            try:
                memo_wif = self.blockchain.wallet.getPrivateKeyForPublicKey(memo_to["memo_key"])
                pubkey = memo_from["memo_key"]
            except MissingKeyError:
                try:
                    # if that failed, we assume that we have sent the memo
                    memo_wif = self.blockchain.wallet.getPrivateKeyForPublicKey(
                        memo_from["memo_key"]
                    )
                    pubkey = memo_to["memo_key"]
                except MissingKeyError:
                    # if all fails, raise exception
                    raise MissingKeyError(
                        "Non of the required memo keys are installed!Need any of {}".format(
                            [memo_to["name"], memo_from["name"]]
                        )
                    )

        if not hasattr(self, "chain_prefix"):
            self.chain_prefix = self.blockchain.prefix

        if message[0] == "#" or memo_to is None or memo_from is None:
            return BtsMemo.decode_memo(PrivateKey(memo_wif), message)
        else:
            return BtsMemo.decode_memo_bts(
                PrivateKey(memo_wif), PublicKey(pubkey, prefix=self.chain_prefix), nonce, message
            )

    def decrypt_binary(self, infile, outfile, buffer_size=2048):
        """Decrypt a binary file

        :param str infile: encrypted binary file
        :param str outfile: output file name
        :param int buffer_size: read buffer size
        :returns: encrypted memo information
        :rtype: dict
        """
        if not os.path.exists(infile):
            raise ValueError("%s does not exists!" % infile)
        if buffer_size % 16 != 0:
            raise ValueError("buffer_size must be dividable by 16")
        with open(infile, "rb") as fin:
            memo_size = struct.unpack("<Q", fin.read(struct.calcsize("<Q")))[0]
            memo = fin.read(memo_size)
            orig_file_size = struct.unpack("<Q", fin.read(struct.calcsize("<Q")))[0]
        memo = "#" + base58encode(hexlify(memo).decode("ascii"))
        memo_to = self.to_account
        memo_from = self.from_account
        from_key, to_key, nonce, check, cipher = BtsMemo.extract_memo_data(memo)
        if memo_to is None and memo_from is None:
            try:
                memo_wif = self.blockchain.wallet.getPrivateKeyForPublicKey(str(to_key))
                pubkey = from_key
            except MissingKeyError:
                try:
                    # if that failed, we assume that we have sent the memo
                    memo_wif = self.blockchain.wallet.getPrivateKeyForPublicKey(str(from_key))
                    pubkey = to_key
                except MissingKeyError:
                    # if all fails, raise exception
                    raise MissingKeyError(
                        "Non of the required memo keys are installed!Need any of {}".format(
                            [str(to_key), str(from_key)]
                        )
                    )
        elif memo_to is not None and memo_from is not None and isinstance(memo_from, PrivateKey):
            memo_wif = memo_from
            if isinstance(memo_to, Account):
                pubkey = memo_to["memo_key"]
            else:
                pubkey = memo_to
        elif memo_to is not None and memo_from is not None and isinstance(memo_to, PrivateKey):
            memo_wif = memo_to
            if isinstance(memo_from, Account):
                pubkey = memo_from["memo_key"]
            else:
                pubkey = memo_from
        else:
            try:
                memo_wif = self.blockchain.wallet.getPrivateKeyForPublicKey(memo_to["memo_key"])
                pubkey = memo_from["memo_key"]
            except MissingKeyError:
                try:
                    # if that failed, we assume that we have sent the memo
                    memo_wif = self.blockchain.wallet.getPrivateKeyForPublicKey(
                        memo_from["memo_key"]
                    )
                    pubkey = memo_to["memo_key"]
                except MissingKeyError:
                    # if all fails, raise exception
                    raise MissingKeyError(
                        "Non of the required memo keys are installed!Need any of {}".format(
                            [memo_to["name"], memo_from["name"]]
                        )
                    )

        if not hasattr(self, "chain_prefix"):
            self.chain_prefix = self.blockchain.prefix
        priv = PrivateKey(memo_wif)
        pubkey = PublicKey(pubkey, prefix=self.chain_prefix)
        nectar_version = BtsMemo.decode_memo(priv, memo)
        shared_secret = BtsMemo.get_shared_secret(priv, pubkey)
        # Init encryption
        aes, checksum = BtsMemo.init_aes(shared_secret, nonce)
        with open(infile, "rb") as fin:
            memo_size = struct.unpack("<Q", fin.read(struct.calcsize("<Q")))[0]
            memo = fin.read(memo_size)
            file_size = struct.unpack("<Q", fin.read(struct.calcsize("<Q")))[0]
            with open(outfile, "wb") as fout:
                while True:
                    data = fin.read(buffer_size)
                    n = len(data)
                    if n == 0:
                        break
                    decd = aes.decrypt(data)
                    n = len(decd)
                    if file_size > n:
                        fout.write(decd)
                    else:
                        fout.write(decd[:file_size])  # <- remove padding on last block
                    file_size -= n
        return {
            "file_size": orig_file_size,
            "from_key": str(from_key),
            "to_key": str(to_key),
            "nonce": nonce,
            "nectar_version": nectar_version,
        }



from __future__ import print_function

import logging
from datetime import timedelta

from nectar import exceptions
from nectar.account import Account
from nectar.blockchain import Blockchain
from nectar.comment import Comment
from nectar.utils import construct_authorperm

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class WatchingTheWatchers:
    def __init__(self):
        self.dvcount = 0
        self.account_type = dict()
        self.account_relations = dict()
        self.by_voter = dict()
        self.by_target = dict()
        self.by_pair = dict()

    def update(self, downvoter, downvoted, dvpower, flagpower):
        pair = downvoter + "/" + downvoted
        if downvoter not in self.by_voter:
            self.by_voter[downvoter] = [0.0, 0.0]
        if downvoted not in self.by_target:
            self.by_target[downvoted] = [0.0, 0.0]
        if pair not in self.by_pair:
            self.by_pair[pair] = [0.0, 0.0]
        self.by_voter[downvoter][0] = self.by_voter[downvoter][0] + dvpower
        self.by_voter[downvoter][1] = self.by_voter[downvoter][1] + flagpower
        self.by_target[downvoted][0] = self.by_target[downvoted][0] + dvpower
        self.by_target[downvoted][1] = self.by_target[downvoted][1] + flagpower
        self.by_pair[pair][0] = self.by_pair[pair][0] + dvpower
        self.by_pair[pair][1] = self.by_pair[pair][1] + flagpower
        self.dvcount = self.dvcount + 1
        if self.dvcount % 100 == 0:
            print(self.dvcount, "downvotes so far.")

    def set_account_info(self, account, fish, related):
        self.account_type[account] = fish
        if len(related) > 0:
            self.account_relations[account] = related

    def report(self):
        print("[REPORT]")
        print(" * account_type :", self.account_type)
        print()
        print(" * account_relations :", self.account_relations)
        print()
        print(" * by voter :", self.by_voter)
        print()
        print(" * by target :", self.by_target)
        print()
        print(" * by pair :", self.by_pair)
        print()
        self.dvcount = 0
        self.account_type = dict()
        self.account_relations = dict()
        self.by_voter = dict()
        self.by_target = dict()
        self.by_pair = dict()


class WatchingTheWatchersBot:
    def __init__(self, wtw):
        self.stopped = None
        self.wtw = wtw
        self.looked_up = set()

    def vote(self, vote_event):
        def process_vote_content(event):
            start_rshares = 0.0
            for vote in event["active_votes"]:
                if vote["voter"] == vote_event["voter"] and float(vote["rshares"]) < 0:
                    if start_rshares + float(vote["rshares"]) < 0:
                        flag_power = 0 - start_rshares - float(vote["rshares"])
                    else:
                        flag_power = 0
                    downvote_power = 0 - vote["rshares"] - flag_power
                    self.wtw.update(vote["voter"], vote_event["author"], downvote_power, flag_power)

        def lookup_accounts(acclist):
            def user_info(accounts):
                if len(acclist) != len(accounts):
                    print("OOPS:", len(acclist), len(accounts), acclist)
                for index in range(0, len(accounts)):
                    a = accounts[index]
                    account = acclist[index]
                    vp = (
                        a["vesting_shares"].amount
                        + a["received_vesting_shares"].amount
                        - a["delegated_vesting_shares"].amount
                    ) / 1000000.0
                    fish = "redfish"
                    if vp >= 1.0:
                        fish = "minnow"
                    if vp >= 10.0:
                        fish = "dolphin"
                    if vp >= 100:
                        fish = "orca"
                    if vp > 1000:
                        fish = "whale"
                    racc = None
                    proxy = None
                    related = list()
                    if a["recovery_account"] != "steem" and a["recovery_account"] != "":
                        related.append(a["recovery_account"])
                    if a["proxy"] != "":
                        related.append(a["proxy"])
                    self.wtw.set_account_info(account, fish, related)
                    accl2 = list()
                    if racc is not None and racc not in self.looked_up:
                        accl2.append(racc)
                    if proxy is not None and proxy not in self.looked_up:
                        accl2.append(proxy)
                    if len(accl2) > 0:
                        lookup_accounts(accl2)

            accounts = []
            for a in acclist:
                accounts.append(Account(a))
            user_info(accounts)

        if vote_event["weight"] < 0:
            authorperm = construct_authorperm(vote_event["author"], vote_event["permlink"])
            # print(authorperm)
            try:
                process_vote_content(Comment(authorperm))
            except exceptions.ContentDoesNotExistsException:
                print("Could not find Comment: %s" % (authorperm))
            al = list()
            if vote_event["voter"] not in self.looked_up:
                al.append(vote_event["voter"])
                self.looked_up.add(vote_event["voter"])
            if vote_event["author"] not in self.looked_up:
                al.append(vote_event["author"])
                self.looked_up.add(vote_event["author"])
            if len(al) > 0:
                lookup_accounts(al)


if __name__ == "__main__":
    wtw = WatchingTheWatchers()
    tb = WatchingTheWatchersBot(wtw)
    blockchain = Blockchain()
    threading = True
    thread_num = 16
    cur_block = blockchain.get_current_block()
    stop = cur_block.identifier
    startdate = cur_block.time() - timedelta(days=1)
    start = blockchain.get_estimated_block_num(startdate, accurate=True)
    for vote in blockchain.stream(
        opNames=["vote"], start=start, stop=stop, threading=threading, thread_num=thread_num
    ):
        tb.vote(vote)
    wtw.report()



# -*- coding: utf-8 -*-
import json
from datetime import date, datetime, timezone

from prettytable import PrettyTable

from nectarbase import operations

from .account import Account
from .amount import Amount
from .blockchainobject import BlockchainObject
from .exceptions import WitnessDoesNotExistsException
from .instance import shared_blockchain_instance
from .utils import formatTimeString


class Witness(BlockchainObject):
    """Read data about a witness in the chain

    :param str owner: Name of the witness
    :param bool lazy: Use lazy loading
    :param bool full: Get full data about witness
    :param nectar.nectar.nectar blockchain_instance: nectar
        instance to use when accesing a RPC
    """

    type_id = 3

    def __init__(self, owner, full=False, lazy=False, blockchain_instance=None, **kwargs):
        self.full = full
        self.lazy = lazy
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        if isinstance(owner, dict):
            owner = self._parse_json_data(owner)
        super(Witness, self).__init__(
            owner, lazy=lazy, full=full, id_item="owner", blockchain_instance=blockchain_instance
        )

    def refresh(self):
        if not self.identifier:
            return
        if not self.blockchain.is_connected():
            return
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        if self.blockchain.rpc.get_use_appbase():
            witness = self.blockchain.rpc.find_witnesses(
                {"owners": [self.identifier]}, api="database"
            )["witnesses"]
            if len(witness) > 0:
                witness = witness[0]
        else:
            witness = self.blockchain.rpc.get_witness_by_account(self.identifier)
        if not witness:
            raise WitnessDoesNotExistsException(self.identifier)
        witness = self._parse_json_data(witness)
        super(Witness, self).__init__(
            witness,
            id_item="owner",
            lazy=self.lazy,
            full=self.full,
            blockchain_instance=self.blockchain,
        )

    def _parse_json_data(self, witness):
        parse_times = [
            "created",
            "last_sbd_exchange_update",
            "hardfork_time_vote",
            "last_hbd_exchange_update",
        ]
        for p in parse_times:
            if p in witness and isinstance(witness.get(p), str):
                witness[p] = formatTimeString(witness.get(p, "1970-01-01T00:00:00"))
        parse_int = [
            "votes",
            "virtual_last_update",
            "virtual_position",
            "virtual_scheduled_time",
        ]
        for p in parse_int:
            if p in witness and isinstance(witness.get(p), str):
                witness[p] = int(witness.get(p, "0"))
        return witness

    def json(self):
        output = self.copy()
        parse_times = [
            "created",
            "last_sbd_exchange_update",
            "hardfork_time_vote",
            "last_hbd_exchange_update",
        ]
        for p in parse_times:
            if p in output:
                p_date = output.get(p, datetime(1970, 1, 1, 0, 0))
                if isinstance(p_date, (datetime, date)):
                    output[p] = formatTimeString(p_date)
                else:
                    output[p] = p_date
        parse_int = [
            "votes",
            "virtual_last_update",
            "virtual_position",
            "virtual_scheduled_time",
        ]
        for p in parse_int:
            if p in output and isinstance(output[p], int):
                output[p] = str(output[p])
        return json.loads(str(json.dumps(output)))

    @property
    def account(self):
        return Account(self["owner"], blockchain_instance=self.blockchain)

    @property
    def is_active(self):
        return (
            len(self["signing_key"]) > 3
            and self["signing_key"][3:] != "1111111111111111111111111111111114T1Anm"
        )

    def feed_publish(self, base, quote=None, account=None):
        """Publish a feed price as a witness.

        :param float base: USD Price of STEEM in SBD (implied price)
        :param float quote: (optional) Quote Price. Should be 1.000 (default), unless
            we are adjusting the feed to support the peg.
        :param str account: (optional) the source account for the transfer
            if not self["owner"]
        """
        quote = quote if quote is not None else "1.000 %s" % (self.blockchain.token_symbol)
        if not account:
            account = self["owner"]
        if not account:
            raise ValueError("You need to provide an account")

        account = Account(account, blockchain_instance=self.blockchain)
        if isinstance(base, Amount):
            base = Amount(base, blockchain_instance=self.blockchain)
        elif isinstance(base, str):
            base = Amount(base, blockchain_instance=self.blockchain)
        else:
            base = Amount(
                base, self.blockchain.backed_token_symbol, blockchain_instance=self.blockchain
            )

        if isinstance(quote, Amount):
            quote = Amount(quote, blockchain_instance=self.blockchain)
        elif isinstance(quote, str):
            quote = Amount(quote, blockchain_instance=self.blockchain)
        else:
            quote = Amount(quote, self.blockchain.token_symbol, blockchain_instance=self.blockchain)

        if not base.symbol == self.blockchain.backed_token_symbol:
            raise AssertionError()
        if not quote.symbol == self.blockchain.token_symbol:
            raise AssertionError()
        op = operations.Feed_publish(
            **{
                "publisher": account["name"],
                "exchange_rate": {
                    "base": base,
                    "quote": quote,
                },
                "prefix": self.blockchain.prefix,
                "json_str": not bool(self.blockchain.config["use_condenser"]),
            }
        )
        return self.blockchain.finalizeOp(op, account, "active")

    def update(self, signing_key, url, props, account=None):
        """Update witness

        :param str signing_key: Signing key
        :param str url: URL
        :param dict props: Properties
        :param str account: (optional) witness account name

        Properties:::

            {
                "account_creation_fee": x,
                "maximum_block_size": x,
                "sbd_interest_rate": x,
            }

        """
        if not account:
            account = self["owner"]
        return self.blockchain.witness_update(signing_key, url, props, account=account)


class WitnessesObject(list):
    def printAsTable(self, sort_key="votes", reverse=True, return_str=False, **kwargs):
        no_feed = False
        if (
            len(self) > 0
            and "sbd_exchange_rate" not in self[0]
            and "hbd_exchange_rate" not in self[0]
        ):
            table_header = ["Name", "Votes [PV]", "Disabled", "Missed", "Fee", "Size", "Version"]
            no_feed = True
        else:
            table_header = [
                "Name",
                "Votes [PV]",
                "Disabled",
                "Missed",
                "Feed base",
                "Feed quote",
                "Feed update",
                "Fee",
                "Size",
                "Interest",
                "Version",
            ]
        if "sbd_exchange_rate" in self[0]:
            bd_exchange_rate = "sbd_exchange_rate"
            bd_interest_rate = "sbd_interest_rate"
            last_bd_exchange_update = "last_sbd_exchange_update"
        else:
            bd_exchange_rate = "hbd_exchange_rate"
            bd_interest_rate = "hbd_interest_rate"
            last_bd_exchange_update = "last_hbd_exchange_update"
        t = PrettyTable(table_header)
        t.align = "l"
        if sort_key == "base":
            sortedList = sorted(
                self, key=lambda self: self[bd_exchange_rate]["base"], reverse=reverse
            )
        elif sort_key == "quote":
            sortedList = sorted(
                self, key=lambda self: self[bd_exchange_rate]["quote"], reverse=reverse
            )
        elif sort_key == "last_sbd_exchange_update" or sort_key == "last_hbd_exchange_update":
            sortedList = sorted(
                self,
                key=lambda self: (
                    datetime.now(timezone.utc) - self[last_bd_exchange_update]
                ).total_seconds(),
                reverse=reverse,
            )
        elif sort_key == "account_creation_fee":
            sortedList = sorted(
                self, key=lambda self: self["props"]["account_creation_fee"], reverse=reverse
            )
        elif sort_key == "sbd_interest_rate" or sort_key == "hbd_interest_rate":
            sortedList = sorted(
                self, key=lambda self: self["props"][bd_interest_rate], reverse=reverse
            )
        elif sort_key == "maximum_block_size":
            sortedList = sorted(
                self, key=lambda self: self["props"]["maximum_block_size"], reverse=reverse
            )
        elif sort_key == "votes":
            sortedList = sorted(self, key=lambda self: int(self[sort_key]), reverse=reverse)
        else:
            sortedList = sorted(self, key=lambda self: self[sort_key], reverse=reverse)
        for witness in sortedList:
            disabled = ""
            if not witness.is_active:
                disabled = "yes"

            if no_feed:
                t.add_row(
                    [
                        witness["owner"],
                        str(round(int(witness["votes"]) / 1e15, 2)),
                        disabled,
                        str(witness["total_missed"]),
                        str(witness["props"]["account_creation_fee"]),
                        str(witness["props"]["maximum_block_size"]),
                        witness["running_version"],
                    ]
                )
            else:
                td = datetime.now(timezone.utc) - witness[last_bd_exchange_update]
                t.add_row(
                    [
                        witness["owner"],
                        str(round(int(witness["votes"]) / 1e15, 2)),
                        disabled,
                        str(witness["total_missed"]),
                        str(
                            Amount(
                                witness[bd_exchange_rate]["base"],
                                blockchain_instance=self.blockchain,
                            )
                        ),
                        str(
                            Amount(
                                witness[bd_exchange_rate]["quote"],
                                blockchain_instance=self.blockchain,
                            )
                        ),
                        str(td.days)
                        + " days "
                        + str(td.seconds // 3600)
                        + ":"
                        + str((td.seconds // 60) % 60),
                        str(
                            Amount(
                                witness["props"]["account_creation_fee"],
                                blockchain_instance=self.blockchain,
                            )
                        ),
                        str(witness["props"]["maximum_block_size"]),
                        str(witness["props"][bd_interest_rate] / 100) + " %",
                        witness["running_version"],
                    ]
                )
        if return_str:
            return t.get_string(**kwargs)
        else:
            print(t.get_string(**kwargs))

    def get_votes_sum(self):
        vote_sum = 0
        for witness in self:
            vote_sum += int(witness["votes"])
        return vote_sum

    def __contains__(self, item):
        from .account import Account

        if isinstance(item, Account):
            name = item["name"]
        elif self.blockchain:
            account = Account(item, blockchain_instance=self.blockchain)
            name = account["name"]

        return any([name == x["owner"] for x in self])

    def __str__(self):
        return self.printAsTable(return_str=True)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, str(self.identifier))


class GetWitnesses(WitnessesObject):
    """Obtain a list of witnesses

    :param list name_list: list of witneses to fetch
    :param int batch_limit: (optional) maximum number of witnesses
        to fetch per call, defaults to 100
    :param Steem steem_instance: Steem() instance to use when
        accessing a RPCcreator = Witness(creator, steem_instance=self)

    .. code-block:: python

        from nectar.witness import GetWitnesses
        w = GetWitnesses(["gtg", "jesta"])
        print(w[0].json())
        print(w[1].json())

    """

    def __init__(
        self, name_list, batch_limit=100, lazy=False, full=True, blockchain_instance=None, **kwargs
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        if not self.blockchain.is_connected():
            return
        witnesses = []
        name_cnt = 0
        if self.blockchain.rpc.get_use_appbase():
            while name_cnt < len(name_list):
                self.blockchain.rpc.set_next_node_on_empty_reply(False)
                witnesses += self.blockchain.rpc.find_witnesses(
                    {"owners": name_list[name_cnt : batch_limit + name_cnt]}, api="database"
                )["witnesses"]
                name_cnt += batch_limit
        else:
            for witness in name_list:
                witnesses.append(self.blockchain.rpc.get_witness_by_account(witness))
        self.identifier = ""
        super(GetWitnesses, self).__init__(
            [
                Witness(x, lazy=lazy, full=full, blockchain_instance=self.blockchain)
                for x in witnesses
            ]
        )


class Witnesses(WitnessesObject):
    """Obtain a list of **active** witnesses and the current schedule

    :param Steem steem_instance: Steem instance to use when
        accesing a RPC

    .. code-block:: python

       >>> from nectar.witness import Witnesses
       >>> Witnesses()
       <Witnesses >

    """

    def __init__(self, lazy=False, full=True, blockchain_instance=None, **kwargs):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        self.lazy = lazy
        self.full = full
        self.refresh()

    def refresh(self):
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        if self.blockchain.rpc.get_use_appbase():
            self.active_witnessess = self.blockchain.rpc.get_active_witnesses(api="database")[
                "witnesses"
            ]
            self.schedule = self.blockchain.rpc.get_witness_schedule(api="database")
            if self.blockchain.config["use_condenser"]:
                self.witness_count = self.blockchain.rpc.get_witness_count(api="condenser")
            else:
                self.witness_count = self.blockchain.rpc.get_witness_count()
        else:
            self.active_witnessess = self.blockchain.rpc.get_active_witnesses()
            self.schedule = self.blockchain.rpc.get_witness_schedule()
            self.witness_count = self.blockchain.rpc.get_witness_count()
        self.current_witness = self.blockchain.get_dynamic_global_properties(use_stored_data=False)[
            "current_witness"
        ]
        self.identifier = ""
        super(Witnesses, self).__init__(
            [
                Witness(x, lazy=self.lazy, full=self.full, blockchain_instance=self.blockchain)
                for x in self.active_witnessess
            ]
        )


class WitnessesVotedByAccount(WitnessesObject):
    """Obtain a list of witnesses which have been voted by an account

    :param str account: Account name
    :param Steem steem_instance: Steem instance to use when
        accesing a RPC

    .. code-block:: python

       >>> from nectar.witness import WitnessesVotedByAccount
       >>> WitnessesVotedByAccount("gtg")
       <WitnessesVotedByAccount gtg>

    """

    def __init__(self, account, lazy=False, full=True, blockchain_instance=None, **kwargs):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        self.account = Account(account, full=True, blockchain_instance=self.blockchain)
        account_name = self.account["name"]
        self.identifier = account_name
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        if self.blockchain.rpc.get_use_appbase():
            if "witnesses_voted_for" not in self.account:
                return
            limit = self.account["witnesses_voted_for"]
            witnessess_dict = self.blockchain.rpc.list_witness_votes(
                {"start": [account_name], "limit": limit, "order": "by_account_witness"},
                api="database",
            )["votes"]
            witnessess = []
            for w in witnessess_dict:
                witnessess.append(w["witness"])
        else:
            if "witness_votes" not in self.account:
                return
            witnessess = self.account["witness_votes"]

        super(WitnessesVotedByAccount, self).__init__(
            [
                Witness(x, lazy=lazy, full=full, blockchain_instance=self.blockchain)
                for x in witnessess
            ]
        )


class WitnessesRankedByVote(WitnessesObject):
    """Obtain a list of witnesses ranked by Vote

    :param str from_account: Witness name from which the lists starts (default = "")
    :param int limit: Limits the number of shown witnesses (default = 100)
    :param Steem steem_instance: Steem instance to use when
        accesing a RPC

    .. code-block:: python

       >>> from nectar.witness import WitnessesRankedByVote
       >>> WitnessesRankedByVote(limit=100)
       <WitnessesRankedByVote >

    """

    def __init__(
        self, from_account="", limit=100, lazy=False, full=False, blockchain_instance=None, **kwargs
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        witnessList = []
        last_limit = limit
        self.identifier = ""
        use_condenser = self.blockchain.config["use_condenser"]
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        if self.blockchain.rpc.get_use_appbase() and not use_condenser:
            query_limit = 1000
        else:
            query_limit = 100
        if self.blockchain.rpc.get_use_appbase() and not use_condenser and from_account == "":
            last_account = None
        elif self.blockchain.rpc.get_use_appbase() and not use_condenser:
            last_account = Witness(from_account, blockchain_instance=self.blockchain)["votes"]
        else:
            last_account = from_account
        if limit > query_limit:
            while last_limit > query_limit:
                tmpList = WitnessesRankedByVote(last_account, query_limit)
                if last_limit < limit:
                    witnessList.extend(tmpList[1:])
                    last_limit -= query_limit - 1
                else:
                    witnessList.extend(tmpList)
                    last_limit -= query_limit
                if self.blockchain.rpc.get_use_appbase():
                    last_account = witnessList[-1]["votes"]
                else:
                    last_account = witnessList[-1]["owner"]
        if last_limit < limit:
            last_limit += 1
        if self.blockchain.rpc.get_use_appbase() and not use_condenser:
            witnessess = self.blockchain.rpc.list_witnesses(
                {"start": [0, last_account], "limit": last_limit, "order": "by_vote_name"},
                api="database",
            )["witnesses"]
        elif self.blockchain.rpc.get_use_appbase() and use_condenser:
            witnessess = self.blockchain.rpc.get_witnesses_by_vote(
                last_account, last_limit, api="condenser"
            )
        else:
            witnessess = self.blockchain.rpc.get_witnesses_by_vote(last_account, last_limit)
        # self.witness_count = len(self.voted_witnessess)
        if last_limit < limit:
            witnessess = witnessess[1:]
        if len(witnessess) > 0:
            for x in witnessess:
                witnessList.append(
                    Witness(x, lazy=lazy, full=full, blockchain_instance=self.blockchain)
                )
        if len(witnessList) == 0:
            return
        super(WitnessesRankedByVote, self).__init__(witnessList)


class ListWitnesses(WitnessesObject):
    """List witnesses ranked by name

    :param str from_account: Witness name from which the lists starts (default = "")
    :param int limit: Limits the number of shown witnesses (default = 100)
    :param Steem steem_instance: Steem instance to use when
        accesing a RPC

    .. code-block:: python

       >>> from nectar.witness import ListWitnesses
       >>> ListWitnesses(from_account="gtg", limit=100)
       <ListWitnesses gtg>

    """

    def __init__(
        self, from_account="", limit=100, lazy=False, full=False, blockchain_instance=None, **kwargs
    ):
        if blockchain_instance is None:
            if kwargs.get("steem_instance"):
                blockchain_instance = kwargs["steem_instance"]
            elif kwargs.get("hive_instance"):
                blockchain_instance = kwargs["hive_instance"]
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        self.identifier = from_account
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        if self.blockchain.rpc.get_use_appbase():
            witnessess = self.blockchain.rpc.list_witnesses(
                {"start": from_account, "limit": limit, "order": "by_name"}, api="database"
            )["witnesses"]
        else:
            witnessess = self.blockchain.rpc.lookup_witness_accounts(from_account, limit)
        if len(witnessess) == 0:
            return
        super(ListWitnesses, self).__init__(
            [
                Witness(x, lazy=lazy, full=full, blockchain_instance=self.blockchain)
                for x in witnessess
            ]
        )



# -*- coding: utf-8 -*-
import unittest

from parameterized import parameterized

from nectar import Hive, Steem
from nectar.account import Account
from nectar.amount import Amount
from nectar.block import Block
from nectar.blockchain import Blockchain
from nectar.comment import Comment
from nectar.instance import set_shared_config, set_shared_steem_instance, shared_steem_instance
from nectar.market import Market
from nectar.price import Price
from nectar.transactionbuilder import TransactionBuilder
from nectar.vote import Vote
from nectar.wallet import Wallet
from nectar.witness import Witness
from nectarapi.exceptions import RPCConnection

from .nodes import get_hive_nodes

# Py3 compatibility

core_unit = "STM"


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        stm = Hive(node=get_hive_nodes())
        stm.config.refreshBackup()
        stm.set_default_nodes(["xyz"])
        del stm

        cls.urls = get_hive_nodes()
        cls.bts = Hive(node=cls.urls, nobroadcast=True, num_retries=10)
        set_shared_steem_instance(cls.bts)
        acc = Account("fullnodeupdate", steem_instance=cls.bts)
        comment = Comment(acc.get_blog_entries(limit=5)[1], steem_instance=cls.bts)
        cls.authorperm = comment.authorperm
        votes = comment.get_votes(raw_data=True)
        last_vote = votes[-1]
        cls.authorpermvoter = comment["authorperm"] + "|" + last_vote["voter"]

    @classmethod
    def tearDownClass(cls):
        stm = Hive(node=get_hive_nodes())
        stm.config.recover_with_latest_backup()

    @parameterized.expand([("instance"), ("steem")])
    def test_account(self, node_param):
        if node_param == "instance":
            set_shared_steem_instance(self.bts)
            acc = Account("test")
            self.assertIn(acc.blockchain.rpc.url, self.urls)
            self.assertIn(acc["balance"].blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Account(
                    "test",
                    steem_instance=Steem(node="https://abc.d", autoconnect=False, num_retries=1),
                )
        else:
            set_shared_steem_instance(Steem(node="https://abc.d", autoconnect=False, num_retries=1))
            stm = self.bts
            acc = Account("test", steem_instance=stm)
            self.assertIn(acc.blockchain.rpc.url, self.urls)
            self.assertIn(acc["balance"].blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Account("test")

    @parameterized.expand([("instance"), ("steem")])
    def test_amount(self, node_param):
        if node_param == "instance":
            stm = Steem(node="https://abc.d", autoconnect=False, num_retries=1)
            set_shared_steem_instance(self.bts)
            o = Amount("1 %s" % self.bts.backed_token_symbol)
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Amount("1 %s" % self.bts.backed_token_symbol, steem_instance=stm)
        else:
            set_shared_steem_instance(Steem(node="https://abc.d", autoconnect=False, num_retries=1))
            stm = self.bts
            o = Amount("1 %s" % self.bts.backed_token_symbol, steem_instance=stm)
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Amount("1 %s" % self.bts.backed_token_symbol)

    @parameterized.expand([("instance"), ("steem")])
    def test_block(self, node_param):
        if node_param == "instance":
            set_shared_steem_instance(self.bts)
            o = Block(1)
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Block(
                    1, steem_instance=Steem(node="https://abc.d", autoconnect=False, num_retries=1)
                )
        else:
            set_shared_steem_instance(Steem(node="https://abc.d", autoconnect=False, num_retries=1))
            stm = self.bts
            o = Block(1, steem_instance=stm)
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Block(1)

    @parameterized.expand([("instance"), ("steem")])
    def test_blockchain(self, node_param):
        if node_param == "instance":
            set_shared_steem_instance(self.bts)
            o = Blockchain()
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Blockchain(
                    steem_instance=Steem(node="https://abc.d", autoconnect=False, num_retries=1)
                )
        else:
            set_shared_steem_instance(Steem(node="https://abc.d", autoconnect=False, num_retries=1))
            stm = self.bts
            o = Blockchain(steem_instance=stm)
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Blockchain()

    @parameterized.expand([("instance"), ("steem")])
    def test_comment(self, node_param):
        if node_param == "instance":
            set_shared_steem_instance(self.bts)
            o = Comment(self.authorperm)
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Comment(
                    self.authorperm,
                    steem_instance=Steem(node="https://abc.d", autoconnect=False, num_retries=1),
                )
        else:
            set_shared_steem_instance(Steem(node="https://abc.d", autoconnect=False, num_retries=1))
            stm = self.bts
            o = Comment(self.authorperm, steem_instance=stm)
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Comment(self.authorperm)

    @parameterized.expand([("instance"), ("steem")])
    def test_market(self, node_param):
        if node_param == "instance":
            set_shared_steem_instance(self.bts)
            o = Market()
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Market(steem_instance=Steem(node="https://abc.d", autoconnect=False, num_retries=1))
        else:
            set_shared_steem_instance(Steem(node="https://abc.d", autoconnect=False, num_retries=1))
            stm = self.bts
            o = Market(steem_instance=stm)
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Market()

    @parameterized.expand([("instance"), ("steem")])
    def test_price(self, node_param):
        if node_param == "instance":
            set_shared_steem_instance(self.bts)
            o = Price(10.0, "%s/%s" % (self.bts.token_symbol, self.bts.backed_token_symbol))
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Price(
                    10.0,
                    "%s/%s" % (self.bts.token_symbol, self.bts.backed_token_symbol),
                    steem_instance=Steem(node="https://abc.d", autoconnect=False, num_retries=1),
                )
        else:
            set_shared_steem_instance(Steem(node="https://abc.d", autoconnect=False, num_retries=1))
            stm = self.bts
            o = Price(
                10.0,
                "%s/%s" % (self.bts.token_symbol, self.bts.backed_token_symbol),
                steem_instance=stm,
            )
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Price(10.0, "%s/%s" % (self.bts.token_symbol, self.bts.backed_token_symbol))

    @parameterized.expand([("instance"), ("steem")])
    def test_vote(self, node_param):
        if node_param == "instance":
            set_shared_steem_instance(self.bts)
            o = Vote(self.authorpermvoter)
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Vote(
                    self.authorpermvoter,
                    steem_instance=Steem(node="https://abc.d", autoconnect=False, num_retries=1),
                )
        else:
            set_shared_steem_instance(Steem(node="https://abc.d", autoconnect=False, num_retries=1))
            stm = self.bts
            o = Vote(self.authorpermvoter, steem_instance=stm)
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Vote(self.authorpermvoter)

    @parameterized.expand([("instance"), ("steem")])
    def test_wallet(self, node_param):
        if node_param == "instance":
            set_shared_steem_instance(self.bts)
            o = Wallet()
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                o = Wallet(
                    steem_instance=Steem(node="https://abc.d", autoconnect=False, num_retries=1)
                )
                o.blockchain.get_config()
        else:
            set_shared_steem_instance(Steem(node="https://abc.d", autoconnect=False, num_retries=1))
            stm = self.bts
            o = Wallet(steem_instance=stm)
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                o = Wallet()
                o.blockchain.get_config()

    @parameterized.expand([("instance"), ("steem")])
    def test_witness(self, node_param):
        if node_param == "instance":
            set_shared_steem_instance(self.bts)
            o = Witness("gtg")
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Witness(
                    "gtg",
                    steem_instance=Steem(node="https://abc.d", autoconnect=False, num_retries=1),
                )
        else:
            set_shared_steem_instance(Steem(node="https://abc.d", autoconnect=False, num_retries=1))
            stm = self.bts
            o = Witness("gtg", steem_instance=stm)
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                Witness("gtg")

    @parameterized.expand([("instance"), ("steem")])
    def test_transactionbuilder(self, node_param):
        if node_param == "instance":
            set_shared_steem_instance(self.bts)
            o = TransactionBuilder()
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                o = TransactionBuilder(
                    steem_instance=Steem(node="https://abc.d", autoconnect=False, num_retries=1)
                )
                o.blockchain.get_config()
        else:
            set_shared_steem_instance(Steem(node="https://abc.d", autoconnect=False, num_retries=1))
            stm = self.bts
            o = TransactionBuilder(steem_instance=stm)
            self.assertIn(o.blockchain.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                o = TransactionBuilder()
                o.blockchain.get_config()

    @parameterized.expand([("instance"), ("steem")])
    def test_steem(self, node_param):
        if node_param == "instance":
            set_shared_steem_instance(self.bts)
            o = Steem(node=self.urls)
            o.get_config()
            self.assertIn(o.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                stm = Steem(node="https://abc.d", autoconnect=False, num_retries=1)
                stm.get_config()
        else:
            set_shared_steem_instance(Steem(node="https://abc.d", autoconnect=False, num_retries=1))
            stm = self.bts
            o = stm
            o.get_config()
            self.assertIn(o.rpc.url, self.urls)
            with self.assertRaises(RPCConnection):
                stm = shared_steem_instance()
                stm.get_config()

    def test_config(self):
        set_shared_config({"node": self.urls})
        set_shared_steem_instance(None)
        o = shared_steem_instance()
        self.assertIn(o.rpc.url, self.urls)



from __future__ import absolute_import, division, print_function, unicode_literals

import unittest
from binascii import hexlify
from builtins import range, super

from nectar.amount import Amount
from nectar.steem import Steem
from nectarbase import operations
from nectarbase.ledgertransactions import Ledger_Transaction
from nectarbase.objects import Operation

TEST_AGAINST_CLI_WALLET = False

prefix = "STEEM"
default_prefix = "STM"
ref_block_num = 34843
ref_block_prefix = 1663841413
expiration = "2020-05-10T20:30:57"
path = "48'/13'/0'/0'/0'"


class Testcases(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.stm = Steem(offline=True)

    def doit(self, printWire=False, ops=None):
        if ops is None:
            ops = [Operation(self.op)]
        tx = Ledger_Transaction(
            ref_block_num=ref_block_num,
            ref_block_prefix=ref_block_prefix,
            expiration=expiration,
            operations=ops,
        )
        txWire = hexlify(bytes(tx)).decode("ascii")
        txApdu = tx.build_apdu(path, chain=prefix)
        if printWire:
            print()
            print(txWire)
            print()
        if len(self.cm) > 0:
            self.assertEqual(self.cm, txWire)
        if len(self.apdu) > 0:
            self.assertEqual(len(self.apdu), len(txApdu))
            for i in range(len(txApdu)):
                self.assertEqual(self.apdu[i], hexlify(txApdu[i]))

    def test_Transfer(self):
        self.op = operations.Transfer(
            **{
                "from": "nettybot",
                "to": "netuoso",
                "amount": Amount("0.001 STEEM", steem_instance=self.stm),
                "memo": "",
                "prefix": default_prefix,
            }
        )
        self.apdu = [
            b"d40400007205800000308000000d80000000800000008000000004"
            b"200000000000000000000000000000000000000000000000000000"
            b"00000000000004021b88040485342c6304048164b85e0401010423"
            b"02086e65747479626f74076e6574756f736f010000000000000003"
            b"535445454d000000040100"
        ]
        self.cm = "04021b88040485342c6304048164b85e040101042302086e65747479626f74076e6574756f736f010000000000000003535445454d000000040100"
        self.doit()

    def test_createclaimedaccount(self):
        self.op = operations.Create_claimed_account(
            **{
                "creator": "netuoso",
                "new_account_name": "netuoso2",
                "owner": {
                    "weight_threshold": 1,
                    "account_auths": [],
                    "key_auths": [["STM7QtTRvd1owAh4uGaC6trxjR9M1cpqfi2WfLQed1GbUGPomt9DP", 1]],
                },
                "active": {
                    "weight_threshold": 1,
                    "account_auths": [],
                    "key_auths": [["STM7QtTRvd1owAh4uGaC6trxjR9M1cpqfi2WfLQed1GbUGPomt9DP", 1]],
                },
                "posting": {
                    "weight_threshold": 1,
                    "account_auths": [],
                    "key_auths": [["STM7QtTRvd1owAh4uGaC6trxjR9M1cpqfi2WfLQed1GbUGPomt9DP", 1]],
                },
                "memo_key": "STM7QtTRvd1owAh4uGaC6trxjR9M1cpqfi2WfLQed1GbUGPomt9DP",
                "json_metadata": "{}",
            }
        )
        self.apdu = [
            b"d4040000dd05800000308000000d800000008000000080000000042000000000000000000000000000000000000000000000000000000000000000000402bd8c04045fe26f450404f179a8570401010481b217076e6574756f736f086e6574756f736f32010000000001034c6a518a9b9e9cb8099176854a322c87db6c7e82c47bd5fe68c273ba63a647160100010000000001034c6a518a9b9e9cb8099176854a322c87db6c7e82c47bd5fe68c273ba63a647160100010000000001034c6a518a9b9e9cb8099176854a322c87db6c7e82c47bd5fe68c273ba63a647160100034c6a",
            b"d404800025518a9b9e9cb8099176854a322c87db6c7e82c47bd5fe68c273ba63a6471600000000040100",
        ]

    def test_vote(self):
        self.op = operations.Vote(
            **{
                "voter": "nettybot",
                "author": "jrcornel",
                "permlink": "hive-sitting-back-at-previous-support-levels-is-this-a-buy",
                "weight": 10000,
            }
        )
        self.cm = b"0402528804049ce2ccea04047660b85e040101045000086e65747479626f74086a72636f726e656c3a686976652d73697474696e672d6261636b2d61742d70726576696f75732d737570706f72742d6c6576656c732d69732d746869732d612d6275791027040100"
        self.apdu = [
            b"d40400009f05800000308000000d800000008000000080000000042000000000000000000000000000000000000000000000000000000000000000000402528804049ce2ccea04047660b85e040101045000086e65747479626f74086a72636f726e656c3a686976652d73697474696e672d6261636b2d61742d70726576696f75732d737570706f72742d6c6576656c732d69732d746869732d612d6275791027040100"
        ]

    def test_pubkey(self):
        tx = Ledger_Transaction(
            ref_block_num=ref_block_num,
            ref_block_prefix=ref_block_prefix,
            expiration=expiration,
            operations=[],
        )
        apdu = tx.build_apdu_pubkey()
        self.assertEqual(
            bytes(apdu),
            b"\xd4\x02\x00\x01\x15\x05\x80\x00\x00\x80\x00\x00\r\x80\x00\x00\x00\x80\x00\x00\x00\x80\x00\x00\x00",
        )



# -*- coding: utf-8 -*-
import unittest

from nectar import Hive, Steem
from nectar.instance import set_shared_blockchain_instance
from nectar.nodelist import NodeList


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        nodelist = NodeList()
        cls.bts = Hive(node=nodelist.get_hive_nodes(), nobroadcast=True, num_retries=10)
        set_shared_blockchain_instance(cls.bts)

    def test_get_nodes(self):
        nodelist = NodeList()
        all_nodes = nodelist.get_nodes(
            exclude_limited=False, dev=True, testnet=True, testnetdev=True
        )
        self.assertEqual(len(nodelist) - 16, len(all_nodes))
        https_nodes = nodelist.get_nodes(wss=False)
        self.assertEqual(https_nodes[0][:5], "https")

    def test_hive_nodes(self):
        nodelist = NodeList()
        nodelist.update_nodes()
        hive_nodes = nodelist.get_hive_nodes()
        for node in hive_nodes:
            blockchainobject = Hive(node=node)
            assert blockchainobject.is_hive

    def test_steem_nodes(self):
        nodelist = NodeList()
        nodelist.update_nodes()
        steem_nodes = nodelist.get_steem_nodes()
        for node in steem_nodes:
            blockchainobject = Steem(node=node)
            assert blockchainobject.is_steem

    def test_nodes_update(self):
        nodelist = NodeList()
        all_nodes = nodelist.get_hive_nodes()
        nodelist.update_nodes(blockchain_instance=self.bts)
        nodes = nodelist.get_hive_nodes()
        self.assertIn(nodes[0], all_nodes)

        all_nodes = nodelist.get_steem_nodes()
        nodelist.update_nodes(blockchain_instance=self.bts)
        nodes = nodelist.get_steem_nodes()
        self.assertIn(nodes[0], all_nodes)



#!/usr/bin/python
import sys

# mpl.use('Agg')
# mpl.use('TkAgg')
import matplotlib.pyplot as plt

from nectar.snapshot import AccountSnapshot

if __name__ == "__main__":
    if len(sys.argv) != 2:
        # print("ERROR: command line parameter mismatch!")
        # print("usage: %s [account]" % (sys.argv[0]))
        account = "thecrazygm"
    else:
        account = sys.argv[1]
    acc_snapshot = AccountSnapshot(account)
    acc_snapshot.get_account_history()
    acc_snapshot.build()
    # acc_snapshot.build(only_ops=["producer_reward"])
    # acc_snapshot.build(only_ops=["curation_reward"])
    # acc_snapshot.build(only_ops=["author_reward"])
    acc_snapshot.build_sp_arrays()
    timestamps = acc_snapshot.timestamps
    own_sp = acc_snapshot.own_sp
    eff_sp = acc_snapshot.eff_sp

    plt.figure(figsize=(12, 6))
    opts = {"linestyle": "-", "marker": "."}
    plt.plot_date(timestamps[1:], own_sp[1:], label="Own SP", **opts)
    plt.plot_date(timestamps[1:], eff_sp[1:], label="Effective SP", **opts)
    plt.grid()
    plt.legend()
    plt.title("SP over time - @%s" % (account))
    plt.xlabel("Date")
    plt.ylabel("SteemPower (SP)")
    # plt.show()
    plt.savefig("sp_over_time-%s.png" % (account))

    print("last effective SP: %.1f SP" % (eff_sp[-1]))
    print("last own SP: %.1f SP" % (own_sp[-1]))



# -*- coding: utf-8 -*-
import unittest

from nectar import Steem, exceptions
from nectar.account import Account
from nectar.instance import set_shared_steem_instance
from nectar.witness import Witness

from .nodes import get_hive_nodes


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.bts = Steem(node=get_hive_nodes(), nobroadcast=True, num_retries=10)
        set_shared_steem_instance(cls.bts)

    def test_Account(self):
        with self.assertRaises(exceptions.AccountDoesNotExistsException):
            Account("FOObarNonExisting")

        c = Account("test")
        self.assertEqual(c["name"], "test")
        self.assertIsInstance(c, Account)

    def test_Witness(self):
        with self.assertRaises(exceptions.WitnessDoesNotExistsException):
            Witness("FOObarNonExisting")

        c = Witness("jesta")
        self.assertEqual(c["owner"], "jesta")
        self.assertIsInstance(c.account, Account)



# -*- coding: utf-8 -*-
import unittest
from decimal import Decimal

from nectar import Steem
from nectar.amount import Amount
from nectar.asset import Asset
from nectar.instance import set_shared_blockchain_instance

from .nodes import get_hive_nodes


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.bts = Steem(node=get_hive_nodes(), nobroadcast=True, num_retries=10)
        set_shared_blockchain_instance(cls.bts)
        cls.asset = Asset("HBD")
        cls.symbol = cls.asset["symbol"]
        cls.precision = cls.asset["precision"]
        cls.asset2 = Asset("HIVE")

    def dotest(self, ret, amount, symbol):
        self.assertEqual(float(ret), float(amount))
        self.assertEqual(ret["symbol"], symbol)
        self.assertIsInstance(ret["asset"], dict)
        self.assertIsInstance(ret["amount"], Decimal)

    def test_init(self):
        stm = self.bts
        # String init
        asset = Asset("HBD", blockchain_instance=stm)
        symbol = asset["symbol"]
        precision = asset["precision"]
        amount = Amount("1 {}".format(symbol), blockchain_instance=stm)
        self.dotest(amount, 1, symbol)

        # Amount init
        amount = Amount(amount, blockchain_instance=stm)
        self.dotest(amount, 1, symbol)

        # blockchain dict init
        amount = Amount(
            {"amount": 1 * 10**precision, "asset_id": asset["id"]}, blockchain_instance=stm
        )
        self.dotest(amount, 1, symbol)

        # API dict init
        amount = Amount(
            {"amount": 1.3 * 10**precision, "asset": asset["id"]}, blockchain_instance=stm
        )
        self.dotest(amount, 1.3, symbol)

        # Asset as symbol
        amount = Amount(1.3, Asset("HBD"), blockchain_instance=stm)
        self.dotest(amount, 1.3, symbol)

        # Asset as symbol
        amount = Amount(1.3, symbol, blockchain_instance=stm)
        self.dotest(amount, 1.3, symbol)

        # keyword inits
        amount = Amount(
            amount=1.3, asset=Asset("HBD", blockchain_instance=stm), blockchain_instance=stm
        )
        self.dotest(amount, 1.3, symbol)

        amount = Amount(
            amount=1.3001, asset=Asset("HBD", blockchain_instance=stm), blockchain_instance=stm
        )
        self.dotest(amount, 1.3001, symbol)

        amount = Amount(
            amount=1.3001,
            asset=Asset("HBD", blockchain_instance=stm),
            fixed_point_arithmetic=True,
            blockchain_instance=stm,
        )
        self.dotest(amount, 1.3, symbol)

        # keyword inits
        amount = Amount(
            amount=1.3, asset=dict(Asset("HBD", blockchain_instance=stm)), blockchain_instance=stm
        )
        self.dotest(amount, 1.3, symbol)

        # keyword inits
        amount = Amount(amount=1.3, asset=symbol, blockchain_instance=stm)
        self.dotest(amount, 1.3, symbol)

        amount = Amount(amount=8.190, asset=symbol, blockchain_instance=stm)
        self.dotest(amount, 8.190, symbol)

    def test_copy(self):
        amount = Amount("1", self.symbol)
        self.dotest(amount.copy(), 1, self.symbol)

    def test_properties(self):
        amount = Amount("1", self.symbol)
        self.assertEqual(amount.amount, 1.0)
        self.assertEqual(amount.symbol, self.symbol)
        self.assertIsInstance(amount.asset, Asset)
        self.assertEqual(amount.asset["symbol"], self.symbol)

    def test_tuple(self):
        amount = Amount("1", self.symbol)
        self.assertEqual(amount.tuple(), (1.0, self.symbol))

    def test_json_appbase(self):
        asset = Asset("HBD", blockchain_instance=self.bts)
        amount = Amount("1", asset, new_appbase_format=False, blockchain_instance=self.bts)
        if self.bts.rpc.get_use_appbase():
            self.assertEqual(
                amount.json(), [str(1 * 10**asset.precision), asset.precision, asset.asset]
            )
        else:
            self.assertEqual(amount.json(), "1.000 HBD")

    def test_json_appbase2(self):
        asset = Asset("HBD", blockchain_instance=self.bts)
        amount = Amount("1", asset, new_appbase_format=True, blockchain_instance=self.bts)
        if self.bts.rpc.get_use_appbase():
            self.assertEqual(
                amount.json(),
                {
                    "amount": str(1 * 10**asset.precision),
                    "nai": asset.asset,
                    "precision": asset.precision,
                },
            )
        else:
            self.assertEqual(amount.json(), "1.000 HBD")

    def test_string(self):
        self.assertEqual(str(Amount("10000", self.symbol)), "10000.000 {}".format(self.symbol))

    def test_int(self):
        self.assertEqual(int(Amount("0.9999", self.symbol)), 999)
        self.assertEqual(int(Amount(0.151, self.symbol)), 151)
        self.assertEqual(int(Amount(8.190, self.symbol)), 8190)
        self.assertEqual(int(Amount(round(0.1509, 3), self.symbol)), 151)
        self.assertEqual(int(Amount(round(0.1509, 3), self.asset)), 151)
        self.assertEqual(int(Amount(int(1), self.symbol)), 1000)
        self.assertEqual(int(Amount(amount=round(0.1509, 3), asset=Asset("HBD"))), 151)

    def test_dict(self):
        self.assertEqual(int(Amount({"amount": "150", "nai": "@@000000021", "precision": 3})), 150)

    def test_float(self):
        self.assertEqual(float(Amount("1", self.symbol)), 1.00000)
        self.assertEqual(float(Amount(0.151, self.symbol)), 0.151)
        self.assertEqual(float(Amount(round(0.1509, 3), self.symbol)), 0.151)
        self.assertEqual(float(Amount(8.190, self.symbol)), 8.190)

    def test_plus(self):
        a1 = Amount(1, self.symbol)
        a2 = Amount(2, self.symbol)
        self.dotest(a1 + a2, 3, self.symbol)
        self.dotest(a1 + 2, 3, self.symbol)
        with self.assertRaises(Exception):
            a1 + Amount(1, asset=self.asset2)
        # inline
        a2 = Amount(2, self.symbol)
        a2 += a1
        self.dotest(a2, 3, self.symbol)
        a2 += 5
        self.dotest(a2, 8, self.symbol)
        a2 += Decimal(2)
        self.dotest(a2, 10, self.symbol)
        with self.assertRaises(Exception):
            a1 += Amount(1, asset=self.asset2)

    def test_minus(self):
        a1 = Amount(1, self.symbol)
        a2 = Amount(2, self.symbol)
        self.dotest(a1 - a2, -1, self.symbol)
        self.dotest(a1 - 5, -4, self.symbol)
        with self.assertRaises(Exception):
            a1 - Amount(1, asset=self.asset2)
        # inline
        a2 = Amount(2, self.symbol)
        a2 -= a1
        self.dotest(a2, 1, self.symbol)
        a2 -= 1
        self.dotest(a2, 0, self.symbol)
        self.dotest(a2 - 2, -2, self.symbol)
        with self.assertRaises(Exception):
            a1 -= Amount(1, asset=self.asset2)

    def test_mul(self):
        a1 = Amount(5, self.symbol)
        a2 = Amount(2, self.symbol)
        self.dotest(a1 * a2, 10, self.symbol)
        self.dotest(a1 * 3, 15, self.symbol)
        with self.assertRaises(Exception):
            a1 * Amount(1, asset=self.asset2)
        # inline
        a2 = Amount(2, self.symbol)
        a2 *= 5
        self.dotest(a2, 10, self.symbol)
        a2 = Amount(2, self.symbol)
        a2 *= a1
        self.dotest(a2, 10, self.symbol)
        with self.assertRaises(Exception):
            a1 *= Amount(2, asset=self.asset2)

    def test_div(self):
        a1 = Amount(15, self.symbol)
        self.dotest(a1 / 3, 5, self.symbol)
        self.dotest(a1 // 2, 7, self.symbol)
        with self.assertRaises(Exception):
            a1 / Amount(1, asset=self.asset2)
        # inline
        a2 = a1.copy()
        a2 /= 3
        self.dotest(a2, 5, self.symbol)
        a2 = a1.copy()
        a2 //= 2
        self.dotest(a2, 7, self.symbol)
        with self.assertRaises(Exception):
            a1 *= Amount(2, asset=self.asset2)

    def test_mod(self):
        a1 = Amount(15, self.symbol)
        a2 = Amount(3, self.symbol)
        self.dotest(a1 % 3, 0, self.symbol)
        self.dotest(a1 % a2, 0, self.symbol)
        self.dotest(a1 % 2, 1, self.symbol)
        with self.assertRaises(Exception):
            a1 % Amount(1, asset=self.asset2)
        # inline
        a2 = a1.copy()
        a2 %= 3
        self.dotest(a2, 0, self.symbol)
        with self.assertRaises(Exception):
            a1 %= Amount(2, asset=self.asset2)

    def test_pow(self):
        a1 = Amount(15, self.symbol)
        a2 = Amount(3, self.symbol)
        self.dotest(a1**3, 15**3, self.symbol)
        self.dotest(a1**a2, 15**3, self.symbol)
        self.dotest(a1**2, 15**2, self.symbol)
        with self.assertRaises(Exception):
            a1 ** Amount(1, asset=self.asset2)
        # inline
        a2 = a1.copy()
        a2 **= 3
        self.dotest(a2, 15**3, self.symbol)
        with self.assertRaises(Exception):
            a1 **= Amount(2, asset=self.asset2)

    def test_ltge(self):
        a1 = Amount(1, self.symbol)
        a2 = Amount(2, self.symbol)
        self.assertTrue(a1 < a2)
        self.assertTrue(a2 > a1)
        self.assertTrue(a2 > 1)
        self.assertTrue(a1 < 5)

    def test_leeq(self):
        a1 = Amount(1, self.symbol)
        a2 = Amount(1, self.symbol)
        self.assertTrue(a1 <= a2)
        self.assertTrue(a1 >= a2)
        self.assertTrue(a1 <= 1)
        self.assertTrue(a1 >= 1)
        self.assertTrue(a1 == 1.0001)

    def test_ne(self):
        a1 = Amount(1, self.symbol)
        a2 = Amount(2, self.symbol)
        self.assertTrue(a1 != a2)
        self.assertTrue(a1 != 5)
        a1 = Amount(1, self.symbol)
        a2 = Amount(1, self.symbol)
        self.assertTrue(a1 == a2)
        self.assertTrue(a1 == 1)



# -*- coding: utf-8 -*-
import hashlib
from binascii import hexlify, unhexlify

from nectargraphenebase.base58 import base58decode, base58encode
from nectargraphenebase.types import varintdecode

try:
    from Cryptodome.Cipher import AES
except ImportError:
    try:
        from Crypto.Cipher import AES
    except ImportError:
        raise ImportError("Missing dependency: pyCryptodome")
import struct

from nectargraphenebase.account import PublicKey

from .objects import Memo

default_prefix = "STM"


def get_shared_secret(priv, pub):
    """Derive the share secret between ``priv`` and ``pub``
    :param `Base58` priv: Private Key
    :param `Base58` pub: Public Key
    :return: Shared secret
    :rtype: hex
    The shared secret is generated such that::
        Pub(Alice) * Priv(Bob) = Pub(Bob) * Priv(Alice)
    """
    pub_point = pub.point()
    priv_point = int(repr(priv), 16)
    res = pub_point * priv_point
    res_hex = "%032x" % res.x()
    # Zero padding
    res_hex = "0" * (64 - len(res_hex)) + res_hex
    return res_hex


def init_aes(shared_secret, nonce):
    """Initialize AES instance
    :param hex shared_secret: Shared Secret to use as encryption key
    :param int nonce: Random nonce
    :return: AES instance
    :rtype: AES
    """
    " Shared Secret "
    ss = hashlib.sha512(unhexlify(shared_secret)).digest()
    " Seed "
    seed = bytes(str(nonce), "ascii") + hexlify(ss)
    seed_digest = hexlify(hashlib.sha512(seed).digest()).decode("ascii")
    " AES "
    key = unhexlify(seed_digest[0:64])
    iv = unhexlify(seed_digest[64:96])
    return AES.new(key, AES.MODE_CBC, iv)


def init_aes_bts(shared_secret, nonce):
    """Initialize AES instance
    :param hex shared_secret: Shared Secret to use as encryption key
    :param int nonce: Random nonce
    :return: AES instance
    :rtype: AES
    """
    " Shared Secret "
    ss = hashlib.sha512(unhexlify(shared_secret)).digest()
    " Seed "
    seed = bytes(str(nonce), "ascii") + hexlify(ss)
    seed_digest = hexlify(hashlib.sha512(seed).digest()).decode("ascii")
    " AES "
    key = unhexlify(seed_digest[0:64])
    iv = unhexlify(seed_digest[64:96])
    return AES.new(key, AES.MODE_CBC, iv)


def init_aes2(shared_secret, nonce):
    """Initialize AES instance
    :param hex shared_secret: Shared Secret to use as encryption key
    :param int nonce: Random nonce
    """
    shared_secret = hashlib.sha512(unhexlify(shared_secret)).hexdigest()
    # Seed
    ss = unhexlify(shared_secret)
    n = struct.pack("<Q", int(nonce))
    encryption_key = hashlib.sha512(n + ss).hexdigest()
    # Check'sum'
    check = hashlib.sha256(unhexlify(encryption_key)).digest()
    check = struct.unpack_from("<I", check[:4])[0]
    # AES
    key = unhexlify(encryption_key[0:64])
    iv = unhexlify(encryption_key[64:96])
    return AES.new(key, AES.MODE_CBC, iv), check


def _pad(s, BS):
    numBytes = BS - len(s) % BS
    return s + numBytes * struct.pack("B", numBytes)


def _unpad(s, BS):
    count = s[-1]
    if s[-count::] == count * struct.pack("B", count):
        return s[:-count]
    return s


def encode_memo_bts(priv, pub, nonce, message):
    """Encode a message with a shared secret between Alice and Bob

    :param PrivateKey priv: Private Key (of Alice)
    :param PublicKey pub: Public Key (of Bob)
    :param int nonce: Random nonce
    :param str message: Memo message
    :return: Encrypted message
    :rtype: hex

    """
    shared_secret = get_shared_secret(priv, pub)
    aes = init_aes_bts(shared_secret, nonce)
    " Checksum "
    raw = bytes(message, "utf8")
    checksum = hashlib.sha256(raw).digest()
    raw = checksum[0:4] + raw
    " Padding "
    raw = _pad(raw, 16)
    " Encryption "
    return hexlify(aes.encrypt(raw)).decode("ascii")


def decode_memo_bts(priv, pub, nonce, message):
    """Decode a message with a shared secret between Alice and Bob

    :param PrivateKey priv: Private Key (of Bob)
    :param PublicKey pub: Public Key (of Alice)
    :param int nonce: Nonce used for Encryption
    :param bytes message: Encrypted Memo message
    :return: Decrypted message
    :rtype: str
    :raise ValueError: if message cannot be decoded as valid UTF-8
           string

    """
    shared_secret = get_shared_secret(priv, pub)
    aes = init_aes_bts(shared_secret, nonce)
    " Encryption "
    raw = bytes(message, "ascii")
    cleartext = aes.decrypt(unhexlify(raw))
    " Checksum "
    checksum = cleartext[0:4]
    message = cleartext[4:]
    message = _unpad(message, 16)
    " Verify checksum "
    check = hashlib.sha256(message).digest()[0:4]
    if check != checksum:  # pragma: no cover
        raise ValueError("checksum verification failure")
    return message.decode("utf8")


def encode_memo(priv, pub, nonce, message, **kwargs):
    """Encode a message with a shared secret between Alice and Bob

    :param PrivateKey priv: Private Key (of Alice)
    :param PublicKey pub: Public Key (of Bob)
    :param int nonce: Random nonce
    :param str message: Memo message
    :return: Encrypted message
    :rtype: hex
    """
    shared_secret = get_shared_secret(priv, pub)
    aes, check = init_aes2(shared_secret, nonce)
    " Padding "
    raw = bytes(message, "utf8")
    raw = _pad(raw, 16)
    " Encryption "
    cipher = hexlify(aes.encrypt(raw)).decode("ascii")
    prefix = kwargs.pop("prefix", default_prefix)
    s = {
        "from": format(priv.pubkey, prefix),
        "to": format(pub, prefix),
        "nonce": nonce,
        "check": check,
        "encrypted": cipher,
        "prefix": prefix,
    }
    tx = Memo(**s)
    return "#" + base58encode(hexlify(bytes(tx)).decode("ascii"))


def extract_memo_data(message):
    """Returns the stored pubkey keys, nonce, checksum and encrypted message of a memo"""
    raw = base58decode(message[1:])
    from_key = PublicKey(raw[:66])
    raw = raw[66:]
    to_key = PublicKey(raw[:66])
    raw = raw[66:]
    nonce = str(struct.unpack_from("<Q", unhexlify(raw[:16]))[0])
    raw = raw[16:]
    check = struct.unpack_from("<I", unhexlify(raw[:8]))[0]
    raw = raw[8:]
    cipher = raw
    return from_key, to_key, nonce, check, cipher


def decode_memo(priv, message):
    """Decode a message with a shared secret between Alice and Bob

    :param PrivateKey priv: Private Key (of Bob)
    :param base58encoded message: Encrypted Memo message
    :return: Decrypted message
    :rtype: str
    :raise ValueError: if message cannot be decoded as valid UTF-8
           string
    """
    # decode structure
    from_key, to_key, nonce, check, cipher = extract_memo_data(message)

    if repr(to_key) == repr(priv.pubkey):
        shared_secret = get_shared_secret(priv, from_key)
    elif repr(from_key) == repr(priv.pubkey):
        shared_secret = get_shared_secret(priv, to_key)
    else:
        raise ValueError("Incorrect PrivateKey")

    # Init encryption
    aes, checksum = init_aes2(shared_secret, nonce)
    # Check
    if not check == checksum:
        raise AssertionError("Checksum failure")
    # Encryption
    # remove the varint prefix (FIXME, long messages!)
    numBytes = 16 - len(cipher) % 16
    n = 16 - numBytes
    message = cipher[n:]
    message = aes.decrypt(unhexlify(bytes(message, "ascii")))
    message = _unpad(message, 16)
    n = varintdecode(message)
    if (len(message) - n) > 0 and (len(message) - n) < 8:
        return "#" + message[len(message) - n :].decode("utf8")
    else:
        return "#" + message.decode("utf8")



# -*- coding: utf-8 -*-
import unittest

from nectar import Steem
from nectar.conveyor import Conveyor
from nectar.instance import set_shared_steem_instance

from .nodes import get_steem_nodes

wif = "5Jh1Gtu2j4Yi16TfhoDmg8Qj3ULcgRi7A49JXdfUUTVPkaFaRKz"


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        stm = Steem(node=get_steem_nodes(), nobroadcast=True, num_retries=10, expiration=120)
        set_shared_steem_instance(stm)

    def test_healthcheck(self):
        health = Conveyor().healthcheck()
        self.assertTrue("version" in health)
        self.assertTrue("ok" in health)
        self.assertTrue("date" in health)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import hashlib
import logging
import struct
import time
from binascii import hexlify

import ecdsa

from .account import PrivateKey, PublicKey

log = logging.getLogger(__name__)

SECP256K1_MODULE = None
SECP256K1_AVAILABLE = False
CRYPTOGRAPHY_AVAILABLE = False
GMPY2_MODULE = False
if not SECP256K1_MODULE:
    try:
        import secp256k1prp as secp256k1

        SECP256K1_MODULE = "secp256k1"
        SECP256K1_AVAILABLE = True
    except Exception:
        try:
            import secp256k1

            SECP256K1_MODULE = "secp256k1"
            SECP256K1_AVAILABLE = True
        except ImportError:
            try:
                import cryptography
            except ImportError:
                SECP256K1_MODULE = "ecdsa"
            else:
                SECP256K1_MODULE = "cryptography"
                CRYPTOGRAPHY_AVAILABLE = True

try:
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives.asymmetric import ec
    from cryptography.hazmat.primitives.asymmetric.utils import (
        decode_dss_signature,
        encode_dss_signature,
    )
except ImportError:
    pass

log.debug("Using SECP256K1 module: %s" % SECP256K1_MODULE)


def _is_canonical(sig):
    sig = bytearray(sig)
    return (
        not (int(sig[0]) & 0x80)
        and not (sig[0] == 0 and not (int(sig[1]) & 0x80))
        and not (int(sig[32]) & 0x80)
        and not (sig[32] == 0 and not (int(sig[33]) & 0x80))
    )


def compressedPubkey(pk):
    if SECP256K1_MODULE == "cryptography" and not isinstance(pk, ecdsa.keys.VerifyingKey):
        order = ecdsa.SECP256k1.order
        x = pk.public_numbers().x
        y = pk.public_numbers().y
    else:
        order = pk.curve.generator.order()
        p = pk.pubkey.point
        x = p.x()
        y = p.y()
    x_str = ecdsa.util.number_to_string(x, order)
    return bytes(chr(2 + (y & 1)), "ascii") + x_str


def recover_public_key(digest, signature, i, message=None):
    """Recover the public key from the the signature"""

    # See http: //www.secg.org/download/aid-780/sec1-v2.pdf section 4.1.6 primarily
    curve = ecdsa.SECP256k1.curve
    G = ecdsa.SECP256k1.generator
    order = ecdsa.SECP256k1.order
    yp = i % 2
    r, s = ecdsa.util.sigdecode_string(signature, order)
    # 1.1
    x = r + (i // 2) * order
    # 1.3. This actually calculates for either effectively 02||X or 03||X depending on 'k' instead of always for 02||X as specified.
    # This substitutes for the lack of reversing R later on. -R actually is defined to be just flipping the y-coordinate in the elliptic curve.
    alpha = ((x * x * x) + (curve.a() * x) + curve.b()) % curve.p()
    beta = ecdsa.numbertheory.square_root_mod_prime(alpha, curve.p())
    y = beta if (beta - yp) % 2 == 0 else curve.p() - beta
    # 1.4 Constructor of Point is supposed to check if nR is at infinity.
    R = ecdsa.ellipticcurve.Point(curve, x, y, order)
    # 1.5 Compute e
    e = ecdsa.util.string_to_number(digest)
    # 1.6 Compute Q = r^-1(sR - eG)
    Q = ecdsa.numbertheory.inverse_mod(r, order) * (s * R + (-e % order) * G)

    if SECP256K1_MODULE == "cryptography" and message is not None:
        if not isinstance(message, bytes):
            message = bytes(message, "utf-8")
        sigder = encode_dss_signature(r, s)
        try:
            Q_point = Q.to_affine()
            public_key = ec.EllipticCurvePublicNumbers(
                Q_point.x(), Q_point.y(), ec.SECP256K1()
            ).public_key(default_backend())
        except Exception:
            try:
                public_key = ec.EllipticCurvePublicNumbers(
                    Q._Point__x, Q._Point__y, ec.SECP256K1()
                ).public_key(default_backend())
            except Exception:
                Q_point = Q.to_affine()
                public_key = ec.EllipticCurvePublicNumbers(
                    int(Q_point.x()), int(Q_point.y()), ec.SECP256K1()
                ).public_key(default_backend())
        public_key.verify(sigder, message, ec.ECDSA(hashes.SHA256()))
        return public_key
    else:
        # Not strictly necessary, but let's verify the message for paranoia's sake.
        if not ecdsa.VerifyingKey.from_public_point(Q, curve=ecdsa.SECP256k1).verify_digest(
            signature, digest, sigdecode=ecdsa.util.sigdecode_string
        ):
            return None
        return ecdsa.VerifyingKey.from_public_point(Q, curve=ecdsa.SECP256k1)


def recoverPubkeyParameter(message, digest, signature, pubkey):
    """Use to derive a number that allows to easily recover the
    public key from the signature
    """
    if not isinstance(message, bytes):
        message = bytes(message, "utf-8")
    for i in range(0, 4):
        if SECP256K1_MODULE == "secp256k1":
            sig = pubkey.ecdsa_recoverable_deserialize(signature, i)
            p = secp256k1.PublicKey(pubkey.ecdsa_recover(message, sig))
            if p.serialize() == pubkey.serialize():
                return i
        elif SECP256K1_MODULE == "cryptography" and not isinstance(pubkey, PublicKey):
            p = recover_public_key(digest, signature, i, message)
            p_comp = hexlify(compressedPubkey(p))
            pubkey_comp = hexlify(compressedPubkey(pubkey))
            if p_comp == pubkey_comp:
                return i
        else:  # pragma: no cover
            p = recover_public_key(digest, signature, i)
            p_comp = hexlify(compressedPubkey(p))
            p_string = hexlify(p.to_string())
            if isinstance(pubkey, PublicKey):
                pubkey_string = bytes(repr(pubkey), "latin")
            else:  # pragma: no cover
                pubkey_string = hexlify(pubkey.to_string())
            if p_string == pubkey_string or p_comp == pubkey_string:
                return i
    return None


def sign_message(message, wif, hashfn=hashlib.sha256):
    """Sign a digest with a wif key

    :param str wif: Private key in
    """

    if not isinstance(message, bytes):
        message = bytes(message, "utf-8")

    digest = hashfn(message).digest()
    priv_key = PrivateKey(wif)
    if SECP256K1_MODULE == "secp256k1":
        p = bytes(priv_key)
        ndata = secp256k1.ffi.new("const int *ndata")
        ndata[0] = 0
        while True:
            ndata[0] += 1
            privkey = secp256k1.PrivateKey(p, raw=True)
            sig = secp256k1.ffi.new("secp256k1_ecdsa_recoverable_signature *")
            signed = secp256k1.lib.secp256k1_ecdsa_sign_recoverable(
                privkey.ctx, sig, digest, privkey.private_key, secp256k1.ffi.NULL, ndata
            )
            if not signed == 1:
                raise AssertionError()
            signature, i = privkey.ecdsa_recoverable_serialize(sig)
            if _is_canonical(signature):
                i += 4  # compressed
                i += 27  # compact
                break
    elif SECP256K1_MODULE == "cryptography":
        cnt = 0
        private_key = ec.derive_private_key(
            int(repr(priv_key), 16), ec.SECP256K1(), default_backend()
        )
        public_key = private_key.public_key()
        while True:
            cnt += 1
            if not cnt % 20:
                log.info("Still searching for a canonical signature. Tried %d times already!" % cnt)
            order = ecdsa.SECP256k1.order
            sigder = private_key.sign(message, ec.ECDSA(hashes.SHA256()))
            r, s = decode_dss_signature(sigder)
            signature = ecdsa.util.sigencode_string(r, s, order)
            # Make sure signature is canonical!
            #
            sigder = bytearray(sigder)
            lenR = sigder[3]
            lenS = sigder[5 + lenR]
            if lenR == 32 and lenS == 32:
                # Derive the recovery parameter
                #
                i = recoverPubkeyParameter(message, digest, signature, public_key)
                i += 4  # compressed
                i += 27  # compact
                break
    else:  # pragma: no branch  # pragma: no cover
        cnt = 0
        p = bytes(priv_key)
        sk = ecdsa.SigningKey.from_string(p, curve=ecdsa.SECP256k1)
        while 1:
            cnt += 1
            if not cnt % 20:
                log.info("Still searching for a canonical signature. Tried %d times already!" % cnt)

            # Deterministic k
            #
            k = ecdsa.rfc6979.generate_k(
                sk.curve.generator.order(),
                sk.privkey.secret_multiplier,
                hashlib.sha256,
                hashlib.sha256(
                    digest
                    + struct.pack("d", time.time())  # use the local time to randomize the signature
                ).digest(),
            )

            # Sign message
            #
            sigder = sk.sign_digest(digest, sigencode=ecdsa.util.sigencode_der, k=k)

            # Reformating of signature
            #
            r, s = ecdsa.util.sigdecode_der(sigder, sk.curve.generator.order())
            signature = ecdsa.util.sigencode_string(r, s, sk.curve.generator.order())

            # Make sure signature is canonical!
            #
            sigder = bytearray(sigder)
            lenR = sigder[3]
            lenS = sigder[5 + lenR]
            if lenR == 32 and lenS == 32:
                # Derive the recovery parameter
                #
                i = recoverPubkeyParameter(message, digest, signature, sk.get_verifying_key())
                i += 4  # compressed
                i += 27  # compact
                break

    # pack signature
    #
    sigstr = struct.pack("<B", i)
    sigstr += signature

    return sigstr


def verify_message(message, signature, hashfn=hashlib.sha256, recover_parameter=None):
    if not isinstance(message, bytes):
        message = bytes(message, "utf-8")
    if not isinstance(signature, bytes):
        signature = bytes(signature, "utf-8")
    if not isinstance(message, bytes):
        raise AssertionError()
    if not isinstance(signature, bytes):
        raise AssertionError()
    digest = hashfn(message).digest()
    sig = signature[1:]
    if recover_parameter is None:
        recover_parameter = bytearray(signature)[0] - 4 - 27  # recover parameter only
    if recover_parameter < 0:
        log.info("Could not recover parameter")
        return None

    if SECP256K1_MODULE == "secp256k1":
        ALL_FLAGS = secp256k1.lib.SECP256K1_CONTEXT_VERIFY | secp256k1.lib.SECP256K1_CONTEXT_SIGN
        # Placeholder
        pub = secp256k1.PublicKey(flags=ALL_FLAGS)
        # Recover raw signature
        sig = pub.ecdsa_recoverable_deserialize(sig, recover_parameter)
        # Recover PublicKey
        verifyPub = secp256k1.PublicKey(pub.ecdsa_recover(message, sig))
        # Convert recoverable sig to normal sig
        normalSig = verifyPub.ecdsa_recoverable_convert(sig)
        # Verify
        verifyPub.ecdsa_verify(message, normalSig)
        phex = verifyPub.serialize(compressed=True)
    elif SECP256K1_MODULE == "cryptography":
        p = recover_public_key(digest, sig, recover_parameter, message)
        order = ecdsa.SECP256k1.order
        r, s = ecdsa.util.sigdecode_string(sig, order)
        sigder = encode_dss_signature(r, s)
        p.verify(sigder, message, ec.ECDSA(hashes.SHA256()))
        phex = compressedPubkey(p)
    else:  # pragma: no branch  # pragma: no cover
        p = recover_public_key(digest, sig, recover_parameter)
        # Will throw an exception of not valid
        p.verify_digest(sig, digest, sigdecode=ecdsa.util.sigdecode_string)
        phex = compressedPubkey(p)

    return phex


def tweakaddPubkey(pk, digest256, SECP256K1_MODULE=SECP256K1_MODULE):
    if SECP256K1_MODULE == "secp256k1":
        tmp_key = secp256k1.PublicKey(pubkey=bytes(pk), raw=True)
        new_key = tmp_key.tweak_add(digest256)  # <-- add
        raw_key = hexlify(new_key.serialize()).decode("ascii")
    else:
        raise Exception("Must have secp256k1 for `tweak_add`")
        # raw_key = ecmult(pk, 1, digest256, SECP256K1_MODULE)
    return PublicKey(raw_key, prefix=pk.prefix)



from __future__ import print_function

import logging

from steem import Steem as steemSteem
from steem.account import Account as steemAccount
from steem.post import Post as steemPost

from nectar import Steem
from nectar.account import Account

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


if __name__ == "__main__":
    stm = Steem("https://api.steemit.com")
    nectar_acc = Account("thecrazygm", steem_instance=stm)
    stm2 = steemSteem(nodes=["https://api.steemit.com"])
    steem_acc = steemAccount("thecrazygm", steemd_instance=stm2)

    # profile
    print("nectar_acc.profile  {}".format(nectar_acc.profile))
    print("steem_acc.profile {}".format(steem_acc.profile))
    # sp
    print("nectar_acc.sp  {}".format(nectar_acc.sp))
    print("steem_acc.sp {}".format(steem_acc.sp))
    # rep
    print("nectar_acc.rep  {}".format(nectar_acc.rep))
    print("steem_acc.rep {}".format(steem_acc.rep))
    # balances
    print("nectar_acc.balances  {}".format(nectar_acc.balances))
    print("steem_acc.balances {}".format(steem_acc.balances))
    # get_balances()
    print("nectar_acc.get_balances()  {}".format(nectar_acc.get_balances()))
    print("steem_acc.get_balances() {}".format(steem_acc.get_balances()))
    # reputation()
    print("nectar_acc.get_reputation()  {}".format(nectar_acc.get_reputation()))
    print("steem_acc.reputation() {}".format(steem_acc.reputation()))
    # voting_power()
    print("nectar_acc.get_voting_power()  {}".format(nectar_acc.get_voting_power()))
    print("steem_acc.voting_power() {}".format(steem_acc.voting_power()))
    # get_followers()
    print("nectar_acc.get_followers()  {}".format(nectar_acc.get_followers()))
    print("steem_acc.get_followers() {}".format(steem_acc.get_followers()))
    # get_following()
    print("nectar_acc.get_following()  {}".format(nectar_acc.get_following()))
    print("steem_acc.get_following() {}".format(steem_acc.get_following()))
    # has_voted()
    print(
        "nectar_acc.has_voted()  {}".format(
            nectar_acc.has_voted("@thecrazygm/api-methods-list-for-appbase")
        )
    )
    print(
        "steem_acc.has_voted() {}".format(
            steem_acc.has_voted(steemPost("@thecrazygm/api-methods-list-for-appbase"))
        )
    )
    # curation_stats()
    print("nectar_acc.curation_stats()  {}".format(nectar_acc.curation_stats()))
    print("steem_acc.curation_stats() {}".format(steem_acc.curation_stats()))
    # virtual_op_count
    print("nectar_acc.virtual_op_count()  {}".format(nectar_acc.virtual_op_count()))
    print("steem_acc.virtual_op_count() {}".format(steem_acc.virtual_op_count()))
    # get_account_votes
    print("nectar_acc.get_account_votes()  {}".format(nectar_acc.get_account_votes()))
    print("steem_acc.get_account_votes() {}".format(steem_acc.get_account_votes()))
    # get_withdraw_routes
    print("nectar_acc.get_withdraw_routes()  {}".format(nectar_acc.get_withdraw_routes()))
    print("steem_acc.get_withdraw_routes() {}".format(steem_acc.get_withdraw_routes()))
    # get_conversion_requests
    print("nectar_acc.get_conversion_requests()  {}".format(nectar_acc.get_conversion_requests()))
    print("steem_acc.get_conversion_requests() {}".format(steem_acc.get_conversion_requests()))
    # export
    # history
    nectar_hist = []
    for h in nectar_acc.history(only_ops=["transfer"]):
        nectar_hist.append(h)
        if len(nectar_hist) >= 10:
            break
    steem_hist = []
    for h in steem_acc.history(filter_by="transfer", start=0):
        steem_hist.append(h)
        if len(steem_hist) >= 10:
            break
    print("nectar_acc.history()  {}".format(nectar_hist))
    print("steem_acc.history() {}".format(steem_hist))
    # history_reverse
    nectar_hist = []
    for h in nectar_acc.history_reverse(only_ops=["transfer"]):
        nectar_hist.append(h)
        if len(nectar_hist) >= 10:
            break
    steem_hist = []
    for h in steem_acc.history_reverse(filter_by="transfer"):
        steem_hist.append(h)
        if len(steem_hist) >= 10:
            break
    print("nectar_acc.history_reverse()  {}".format(nectar_hist))
    print("steem_acc.history_reverse() {}".format(steem_hist))



#!/usr/bin/python
import sys

# mpl.use('Agg')
# mpl.use('TkAgg')
import matplotlib.pyplot as plt

from nectar.snapshot import AccountSnapshot

if __name__ == "__main__":
    if len(sys.argv) != 2:
        # print("ERROR: command line parameter mismatch!")
        # print("usage: %s [account]" % (sys.argv[0]))
        account = "thecrazygm"
    else:
        account = sys.argv[1]
    acc_snapshot = AccountSnapshot(account)
    acc_snapshot.get_account_history()
    acc_snapshot.build(enable_rewards=True)
    acc_snapshot.build_curation_arrays()
    timestamps = acc_snapshot.curation_per_1000_SP_timestamp
    curation_per_1000_SP = acc_snapshot.curation_per_1000_SP

    plt.figure(figsize=(12, 6))
    opts = {"linestyle": "-", "marker": "."}
    plt.plot_date(
        timestamps, curation_per_1000_SP, label="Curation reward per week and 1k SP", **opts
    )
    plt.grid()
    plt.legend()
    plt.title("Curation over time - @%s" % (account))
    plt.xlabel("Date")
    plt.ylabel("Curation rewards (SP / (week * 1k SP))")
    plt.show()
    # plt.savefig("curation_per_week-%s.png" % (account))
    print("last curation reward per week and 1k sp %.2f SP" % (curation_per_1000_SP[-1]))
